'''
Unit tests for CurvRect3DRegridder

@author: Karl Smith
'''

from __future__ import print_function

import unittest
import numpy
import ESMP
from esmpcontrol import ESMPControl
from regrid3d import CurvRect3DRegridder


class CurvRect3DRegridderTests(unittest.TestCase):
    '''
    Unit tests for the CurvRect3DRegridder class
    '''

    # flag to indicate when to call ESMPControl().stopESMP()
    last_test = False


    def setUp(self):
        '''
        Create some repeatedly used test data.
        '''
        # Use tuples for the arrays to make sure the NumPy 
        # arrays created in the class methods are always used;
        # not arrays that happened to be passed as input.
        # Also verifies passed data is not modified.

        # Rectilinear coordinates and data
        crn_lons = numpy.linspace(-16.0,  16.0, 17)
        crn_lats = numpy.linspace(-11.0,  11.0, 12)
        crn_levs = numpy.linspace( 10.0, 100.0, 10)
        crn_shape = (crn_lons.shape[0], crn_lats.shape[0], crn_levs.shape[0])

        ctr_lons = 0.5 * (crn_lons[:-1] + crn_lons[1:])
        ctr_lats = 0.5 * (crn_lats[:-1] + crn_lats[1:])
        ctr_levs = 0.5 * (crn_levs[:-1] + crn_levs[1:])
        ctr_shape = (ctr_lons.shape[0], ctr_lats.shape[0], ctr_levs.shape[0])

        # Create in C order so the tuplets are [lon][lat][lev] 
        ctr_lons_mat = numpy.repeat(ctr_lons, ctr_shape[1] * ctr_shape[2]) \
                            .reshape(ctr_shape)
        ctr_lats_mat = numpy.repeat(numpy.tile(ctr_lats, ctr_shape[0]),
                                    ctr_shape[2]) \
                            .reshape(ctr_shape)
        ctr_levs_mat = numpy.tile(ctr_levs, ctr_shape[0] * ctr_shape[1]) \
                            .reshape(ctr_shape)

        data = 2.0 * numpy.cos(numpy.deg2rad(ctr_lons_mat + 20.0)) \
                   * numpy.cos(numpy.deg2rad(ctr_lats_mat + 15.0)) \
                   / numpy.log10(ctr_levs_mat + 10.0)

        ctr_flags = numpy.zeros(ctr_shape, dtype=numpy.int32)
        ctr_flags[ 0,  0, :] = 1
        ctr_flags[ 0, -1, :] = 1
        ctr_flags[-1, -1, :] = 1

        crn_flags = numpy.zeros(crn_shape, dtype=numpy.int32)
        crn_flags[ 0,  0, :] = 1
        crn_flags[ 0, -1, :] = 1
        crn_flags[-1, -1, :] = 1

        self.rect_corner_lons = tuple(crn_lons)
        self.rect_corner_lats = tuple(crn_lats)
        self.rect_corner_levs = tuple(crn_levs)
        self.rect_center_lons = tuple(ctr_lons)
        self.rect_center_lats = tuple(ctr_lats)
        self.rect_center_levs = tuple(ctr_levs)
        self.rect_center_ignr = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in ctr_flags.tolist()])
        self.rect_corner_ignr = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in crn_flags.tolist()])
        self.rect_data        = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in data.tolist()])

        # Curvilinear coordinates and data - expanded out a step on all region sides
        crn_lons = numpy.linspace(-18.0,  18.0, 19)
        crn_lats = numpy.linspace(-13.0,  13.0, 14)
        crn_levs = numpy.linspace(  0.0, 110.0, 12)
        crn_shape = (crn_lons.shape[0], crn_lats.shape[0], crn_levs.shape[0])

        ctr_lons = 0.5 * (crn_lons[:-1] + crn_lons[1:])
        ctr_lats = 0.5 * (crn_lats[:-1] + crn_lats[1:])
        ctr_levs = 0.5 * (crn_levs[:-1] + crn_levs[1:])
        ctr_shape = (ctr_lons.shape[0], ctr_lats.shape[0], ctr_levs.shape[0])

        # Create in C order so the tuplets are [lon][lat][lev] 
        crn_lons_mat = numpy.repeat(crn_lons, crn_shape[1] * crn_shape[2]) \
                            .reshape(crn_shape)
        crn_lats_mat = numpy.repeat(numpy.tile(crn_lats, crn_shape[0]),
                                    crn_shape[2]) \
                            .reshape(crn_shape)
        crn_levs_mat = numpy.tile(crn_levs, crn_shape[0] * crn_shape[1]) \
                            .reshape(crn_shape)

        # Create in C order so the tuplets are [lon][lat][lev] 
        ctr_lons_mat = numpy.repeat(ctr_lons, ctr_shape[1] * ctr_shape[2]) \
                            .reshape(ctr_shape)
        ctr_lats_mat = numpy.repeat(numpy.tile(ctr_lats, ctr_shape[0]),
                                    ctr_shape[2]) \
                            .reshape(ctr_shape)
        ctr_levs_mat = numpy.tile(ctr_levs, ctr_shape[0] * ctr_shape[1]) \
                            .reshape(ctr_shape)

        # Tweak the coordinates some, pulling in near edges of the region
        crn_lons = crn_lons_mat * numpy.cos(numpy.deg2rad(crn_lats_mat))
        crn_lats = crn_lats_mat * numpy.cos(numpy.deg2rad(crn_lons_mat))
        crn_levs = crn_levs_mat * numpy.cos(numpy.deg2rad(crn_lats_mat) / 2.0) \
                                * numpy.cos(numpy.deg2rad(crn_lons_mat) / 2.0) \
                                * numpy.cos(numpy.deg2rad(crn_levs_mat - 55.0) / 5.0)

        ctr_lons = ctr_lons_mat * numpy.cos(numpy.deg2rad(ctr_lats_mat))
        ctr_lats = ctr_lats_mat * numpy.cos(numpy.deg2rad(ctr_lons_mat))
        ctr_levs = ctr_levs_mat * numpy.cos(numpy.deg2rad(ctr_lats_mat) / 2.0) \
                                * numpy.cos(numpy.deg2rad(ctr_lons_mat) / 2.0) \
                                * numpy.cos(numpy.deg2rad(ctr_levs_mat - 55.0) / 5.0)

        data = 2.0 * numpy.cos(numpy.deg2rad(ctr_lons + 20.0)) \
                   * numpy.cos(numpy.deg2rad(ctr_lats + 15.0)) \
                   / numpy.log10(ctr_levs + 10.0)

        ctr_flags = numpy.zeros(ctr_shape, dtype=numpy.int32)
        ctr_flags[ :2,  :2, :] = 1
        ctr_flags[ :2, -2:, :] = 1
        ctr_flags[-2:, -2:, :] = 1

        crn_flags = numpy.zeros(crn_shape, dtype=numpy.int32)
        crn_flags[ :2,  :2, :] = 1
        crn_flags[ :2, -2:, :] = 1
        crn_flags[-2:, -2:, :] = 1
 
        self.curv_corner_lons = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in crn_lons.tolist()])
        self.curv_corner_lats = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in crn_lats.tolist()])
        self.curv_corner_levs = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in crn_levs.tolist()])
        self.curv_center_lons = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in ctr_lons.tolist()])
        self.curv_center_lats = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in ctr_lats.tolist()])
        self.curv_center_levs = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in ctr_levs.tolist()])
        self.curv_center_ignr = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in ctr_flags.tolist()])
        self.curv_corner_ignr = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in crn_flags.tolist()])
        self.curv_data        = tuple([tuple([tuple(subsubarr) for subsubarr in subarr])
                                       for subarr in data.tolist()])

        # undef_val must be a numpy array
        self.undef_val = numpy.array([1.0E10], dtype=numpy.float64)

        if not ESMPControl().startCheckESMP():
            self.fail("startCheckESMP did not succeed - test called after last_test set to True")


    def test01CurvRectRegridderInit(self):
        '''
        Test of the CurvRect3DRegridder.__init__ method.
        '''
        regridder = CurvRect3DRegridder()
        self.assertTrue(regridder != None, "CurvRect3DRegridder() returned None")
        regridder.finalize()


    def test02CreateCurvGrid(self):
        '''
        Tests the CurvRect3DRegridder.createCurvGrid method.
        Since nothing is returned from this method, just
        checks for unexpected/expected Errors being raised.
        '''
        regridder = CurvRect3DRegridder()

        # Test with all corner and center data
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, self.curv_center_ignr,
                                 True,
                                 self.curv_corner_lons, self.curv_corner_lats,
                                 self.curv_corner_levs, self.curv_corner_ignr)

        # Test without flags 
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, None, True,
                                 self.curv_corner_lons, self.curv_corner_lats,
                                 self.curv_corner_levs)

        # Test without corners
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, self.curv_center_ignr)

        # Test without corners or flags
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs)

        # Test invalid cases

        # Done with this regridder
        regridder.finalize()


    def test03AssignCurvField(self):
        '''
        Tests the CurvRect3DRegridder.assignCurvGrid method.
        Since nothing is returned from this method, just
        checks for unexpected/expected Errors being raised.
        '''
        regridder = CurvRect3DRegridder()

        # Test with all corner and center data
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, self.curv_center_ignr,
                                 True,
                                 self.curv_corner_lons, self.curv_corner_lats,
                                 self.curv_corner_levs, self.curv_corner_ignr)
        regridder.assignCurvField()
        regridder.assignCurvField(self.curv_data)

        # Test without flags 
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, None, True,
                                 self.curv_corner_lons, self.curv_corner_lats,
                                 self.curv_corner_levs)
        regridder.assignCurvField(self.curv_data)
        regridder.assignCurvField()

        # Test without corners
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, self.curv_center_ignr)
        regridder.assignCurvField(self.curv_data)
        regridder.assignCurvField()

        # Test without corners or flags
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs)
        regridder.assignCurvField()
        regridder.assignCurvField(self.curv_data)

        # Test invalid cases

        # Done with this regridder
        regridder.finalize()


    def test04CreateRectGrid(self):
        '''
        Tests the CurvRect3DRegridder.createRectGrid method.
        Since nothing is returned from this method, just
        checks for unexpected/expected Errors being raised.
        '''
        regridder = CurvRect3DRegridder()

        # Test with all corner and center data
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, self.rect_center_ignr,
                                 True,
                                 self.rect_corner_lons, self.rect_corner_lats,
                                 self.rect_corner_levs, self.rect_corner_ignr)

        # Test without flags
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, None, True,
                                 self.rect_corner_lons, self.rect_corner_lats,
                                 self.rect_corner_levs)

        # Test without corners
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, self.rect_center_ignr)

        # Test without corners or flags 
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs)

        # Test invalid cases

        # Done with this regridder
        regridder.finalize()


    def test05AssignRectField(self):
        '''
        Tests the CurvRect3DRegridder.assignRectGrid method.
        Since nothing is returned from this method, just
        checks for unexpected/expected Errors being raised.
        '''
        regridder = CurvRect3DRegridder()

        # Test with all corner and center data
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, self.rect_center_ignr,
                                 True,
                                 self.rect_corner_lons, self.rect_corner_lats,
                                 self.rect_corner_levs, self.rect_corner_ignr)
        regridder.assignRectField(self.rect_data)
        regridder.assignRectField()

        # Test without flags
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, None, True,
                                 self.rect_corner_lons, self.rect_corner_lats,
                                 self.rect_corner_levs)
        regridder.assignRectField()
        regridder.assignRectField(self.rect_data)

        # Test without corners
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, self.rect_center_ignr)
        regridder.assignRectField()
        regridder.assignRectField(self.rect_data)

        # Test without corners or flags 
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs)
        regridder.assignRectField(self.rect_data)
        regridder.assignRectField()

        # Test invalid cases

        # Done with this regridder
        regridder.finalize()


    def test06RegridCurvToRectBilinear(self):
        '''
        Tests the CurvRect3DRegridder.regridCurvToRect method.
        '''
        margin = 1
        delta = 1.0
      
        regridder = CurvRect3DRegridder()

        # Test with only center data but no flags
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs)
        regridder.assignCurvField(self.curv_data)
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs)
        regridder.assignRectField()
        regrid_data = regridder.regridCurvToRect(self.undef_val,
                                                 ESMP.ESMP_REGRIDMETHOD_BILINEAR)
        expect_data = numpy.array(self.rect_data, dtype=numpy.float64)
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.rect_center_lons[i], 
                              self.rect_center_lats[j], self.rect_center_levs[k]))
        if mismatch_found:
            self.fail("data mismatch found for bilinear regridding without flags")

        # Test with only center data and with flags
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, self.curv_center_ignr)
        regridder.assignCurvField(self.curv_data)
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, self.rect_center_ignr)
        regridder.assignRectField()
        regrid_data = regridder.regridCurvToRect(self.undef_val,
                                                 ESMP.ESMP_REGRIDMETHOD_BILINEAR)
        expect_data = numpy.array(self.rect_data, dtype=numpy.float64)
        undef_flags = numpy.array(self.rect_center_ignr, dtype=numpy.bool)
        expect_data[undef_flags] = self.undef_val
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.rect_center_lons[i], 
                              self.rect_center_lats[j], self.rect_center_levs[k]))
        if mismatch_found:
            self.fail("data mismatch found for bilinear regridding with flags")


    def test07RegridCurvToRectConserve(self):
        '''
        Tests the CurvRect3DRegridder.regridCurvToRect method.
        '''
        margin = 1
        delta = 1.0
      
        regridder = CurvRect3DRegridder()

        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, None, True,
                                 self.curv_corner_lons, self.curv_corner_lats,
                                 self.curv_corner_levs)
        regridder.assignCurvField(self.curv_data)
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, None, True,
                                 self.rect_corner_lons, self.rect_corner_lats,
                                 self.rect_corner_levs)
        regridder.assignRectField()
        regrid_data = regridder.regridCurvToRect(self.undef_val, 
                                                 ESMP.ESMP_REGRIDMETHOD_CONSERVE)
        expect_data = numpy.array(self.rect_data, dtype=numpy.float64)
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.rect_center_lons[i], 
                              self.rect_center_lats[j], self.rect_center_levs[k]))
        if mismatch_found:
            self.fail("data mismatch found for conservative regridding")



    def test08RegridCurvToRectPatch(self):
        '''
        Tests the CurvRect3DRegridder.regridCurvToRect method.
        '''
        margin = 1
        delta = 1.0
      
        regridder = CurvRect3DRegridder()

        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs)
        regridder.assignCurvField(self.curv_data)
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs)
        regridder.assignRectField()
        regrid_data = regridder.regridCurvToRect(self.undef_val,
                                                 ESMP.ESMP_REGRIDMETHOD_PATCH)
        expect_data = numpy.array(self.rect_data, dtype=numpy.float64)
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.rect_center_lons[i], 
                              self.rect_center_lats[j], self.rect_center_levs[k]))
        if mismatch_found:
            self.fail("data mismatch found for patch regridding")


    def test09RegridRectToCurvBilinear(self):
        '''
        Tests the CurvRect3DRegridder.regridRectToCurv method.
        '''
        margin = 2
        delta = 1.0

        regridder = CurvRect3DRegridder()

        # Test with only center data but no flags
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs)
        regridder.assignCurvField()
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs)
        regridder.assignRectField(self.rect_data)
        regrid_data = regridder.regridRectToCurv(self.undef_val,
                                                 ESMP.ESMP_REGRIDMETHOD_BILINEAR)
        expect_data = numpy.array(self.curv_data, dtype=numpy.float64)
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.curv_center_lons[i][j][k], 
                              self.curv_center_lats[i][j][k], self.curv_center_levs[i][j][k]))
        if mismatch_found:
            self.fail("data mismatch found for bilinear regridding without flags")

        # Test with only center data and with flags
        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, self.curv_center_ignr)
        regridder.assignCurvField()
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, self.rect_center_ignr)
        regridder.assignRectField(self.rect_data)
        regrid_data = regridder.regridRectToCurv(self.undef_val,
                                                 ESMP.ESMP_REGRIDMETHOD_BILINEAR)
        expect_data = numpy.array(self.curv_data, dtype=numpy.float64)
        undef_flags = numpy.array(self.rect_center_ignr, dtype=numpy.bool)
        expect_data[undef_flags] = self.undef_val
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.curv_center_lons[i][j][k], 
                              self.curv_center_lats[i][j][k], self.curv_center_levs[i][j][k]))
        if mismatch_found:
            self.fail("data mismatch found for bilinear regridding with flags")


    def test10RegridRectToCurvConserve(self):
        '''
        Tests the CurvRect3DRegridder.regridRectToCurv method.
        '''
        margin = 2
        delta = 1.0

        regridder = CurvRect3DRegridder()

        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs, None, True,
                                 self.curv_corner_lons, self.curv_corner_lats,
                                 self.curv_corner_levs)
        regridder.assignCurvField()
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs, None, True,
                                 self.rect_corner_lons, self.rect_corner_lats,
                                 self.rect_corner_levs)
        regridder.assignRectField(self.rect_data)
        regrid_data = regridder.regridRectToCurv(self.undef_val,
                                                 ESMP.ESMP_REGRIDMETHOD_CONSERVE)
        expect_data = numpy.array(self.curv_data, dtype=numpy.float64)
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.curv_center_lons[i][j][k], 
                              self.curv_center_lats[i][j][k], self.curv_center_levs[i][j][k]))
        if mismatch_found:
            self.fail("data mismatch found for conservative regridding")


    def test11RegridRectToCurvPatch(self):
        '''
        Tests the CurvRect3DRegridder.regridRectToCurv method.
        '''
        margin = 2
        delta = 1.0

        # Mark as the last test so ESMPControl().stopESMP will be called
        self.last_test = True

        regridder = CurvRect3DRegridder()

        regridder.createCurvGrid(self.curv_center_lons, self.curv_center_lats,
                                 self.curv_center_levs)
        regridder.assignCurvField()
        regridder.createRectGrid(self.rect_center_lons, self.rect_center_lats,
                                 self.rect_center_levs)
        regridder.assignRectField(self.rect_data)
        regrid_data = regridder.regridRectToCurv(self.undef_val,
                                                 ESMP.ESMP_REGRIDMETHOD_PATCH)
        expect_data = numpy.array(self.curv_data, dtype=numpy.float64)
        mismatch_found = False
        for i in range(margin, expect_data.shape[0] - margin):
            for j in range(margin, expect_data.shape[1] - margin):
                for k in range(margin, expect_data.shape[2] - margin):
                    if numpy.abs(expect_data[i, j, k] - regrid_data[i, j, k]) > delta:
                        mismatch_found = True
                        print("expect = %#6.4f, found = %#6.4f for lon = %5.1f, " \
                              "lat = %5.1f, lev = %5.1f" % (expect_data[i, j, k], 
                              regrid_data[i, j, k], self.curv_center_lons[i][j][k], 
                              self.curv_center_lats[i][j][k], self.curv_center_levs[i][j][k]))
        if mismatch_found:
            self.fail("data mismatch found for patch regridding")


    def tearDown(self):
        '''
        Finalize ESMP if it has been initialized and if this is the last test
        '''
        if self.last_test:
            ESMPControl().stopESMP(False)


if __name__ == "__main__":
    '''
    Run the unit tests in this module.
    '''
    unittest.main()




"""
Returns the array of percent point function values for
a probability distribution and set of quantile values.
"""

from __future__ import print_function

import numpy
import scipy.stats
import pyferret
import pyferret.stats


def ferret_init(id):
    """
    Initialization for the stats_ppf python-backed ferret external function
    """
    axes_values = [ pyferret.AXIS_IMPLIED_BY_ARGS ] * pyferret.MAX_FERRET_NDIM
    true_influences = [ True ] * pyferret.MAX_FERRET_NDIM
    false_influences = [ False ] * pyferret.MAX_FERRET_NDIM
    retdict = { "numargs": 3,
                "descript": "Returns percent point function (inverse of cdf) values for a probability distribution",
                "axes": axes_values,
                "argnames": ("PROBS", "PDNAME", "PDPARAMS"),
                "argdescripts": ("Probabilities (0-1) at which to calculate the percent point function values",
                                 "Name of a probability distribution",
                                 "Parameters for this probability distribution"),
                "argtypes": (pyferret.FLOAT_ARRAY, pyferret.STRING_ONEVAL, pyferret.FLOAT_ARRAY),
                "influences": (true_influences, false_influences, false_influences),
              }
    return retdict


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Assigns result with the percent point function values for the probability
    distribution indicated by inputs[1] (a string) using the parameters given
    in inputs[2] at the quantile values given by inputs[0].  For undefined
    quantile values, the result value will be undefined.
    """
    distribname = inputs[1]
    distribparams = inputs[2].reshape(-1)
    distrib = pyferret.stats.getdistrib(distribname, distribparams)
    badmask = ( numpy.fabs(inputs[0] - inpbdfs[0]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(inputs[0]))
    goodmask = numpy.logical_not(badmask)
    result[badmask] = resbdf
    # array[goodmask] is a flattened array
    result[goodmask] = distrib.ppf(inputs[0][goodmask])


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # make sure ferret_init does not have problems
    info = ferret_init(0)

    # Normal distribution along the Y axis
    dimen = 25
    mu = 5.0
    sigma = 2.0
    distf = scipy.stats.norm(mu, sigma)
    qvals = numpy.linspace(0.05, 0.95, dimen)
    ppfvals = distf.ppf(qvals)

    pfname = "norm"
    pfparams = numpy.array([mu, sigma], dtype=numpy.float64)
    inpbdfs = numpy.array([-1.0, 0.0, 0.0], dtype=numpy.float64)
    resbdf = numpy.array([-2.0], dtype=numpy.float64)
    quantile = numpy.empty((1, dimen, 1, 1, 1, 1), dtype=numpy.float64, order='F')
    expected = numpy.empty((1, dimen, 1, 1, 1, 1), dtype=numpy.float64, order='F')
    for j in range(dimen):
        if (j % 7) == 2:
            quantile[0, j, 0, 0, 0, 0] = inpbdfs[0]
            expected[0, j, 0, 0, 0, 0] = resbdf[0]
        else:
            quantile[0, j, 0, 0, 0, 0] = qvals[j]
            expected[0, j, 0, 0, 0, 0] = ppfvals[j]
    result = -888.0 * numpy.ones((1, dimen, 1, 1, 1, 1), dtype=numpy.float64, order='F')
    ferret_compute(0, result, resbdf, (quantile, pfname, pfparams), inpbdfs)
    if not numpy.allclose(result, expected):
        print("Expected (flattened) = %s" % str(expected.reshape(-1)))
        print("Result (flattened) = %s" % str(result.reshape(-1)))
        raise ValueError("Unexpected result")

    # All successful
    print("Success")




"""
Returns the array of standard scores for an array of data.  The
standard score are for the standard distribution centered of the
mean value of the data with the same variance as the data.
"""

from __future__ import print_function

import math
import numpy
import pyferret
import scipy.stats


def ferret_init(id):
    """
    Initialization for the stats_zscore PyEF
    """
    axes_values = [ pyferret.AXIS_IMPLIED_BY_ARGS ] * pyferret.MAX_FERRET_NDIM
    true_influences = [ True ] * pyferret.MAX_FERRET_NDIM
    retdict = { "numargs": 1,
                "descript": "Returns standard scores for data values relative to " \
                            "a normal distribution with same mean and variance as the data",
                "axes": axes_values,
                "argnames": ( "VALUES", ),
                "argdescripts": ( "Array of data values", ),
                "argtypes": ( pyferret.FLOAT_ARRAY, ),
                "influences": ( true_influences, ),
              }
    return retdict


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Assigns result standard scores of data values given in inputs[0]
    relative to a normal distribution with the same mean and variance
    as the data.  For undefined data values, the result value will
    be undefined.
    """
    badmask = ( numpy.fabs(inputs[0] - inpbdfs[0]) < 1.0E-5 )
    badmask = numpy.logical_or(badmask, numpy.isnan(inputs[0]))
    goodmask = numpy.logical_not(badmask)
    result[badmask] = resbdf
    # convert to 64-bit for precision in calculating the mean and variance
    sample = numpy.array(inputs[0][goodmask], dtype=numpy.float64)
    # array[goodmask] is a flattened array
    result[goodmask] = scipy.stats.zscore(sample)


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # make sure ferret_init does not have problems
    info = ferret_init(0)

    # Get a random sample and the expected standard scores
    ydim = 83
    zdim = 17
    samplesize = 1300
    sample = scipy.stats.norm(5.0, 2.0).rvs(samplesize)
    zscores = (sample - sample.mean()) / math.sqrt(sample.var(0))

    # setup for the call to ferret_compute
    inpbdfs = numpy.array([-9999.0], dtype=numpy.float64)
    resbdf = numpy.array([-8888.0], dtype=numpy.float64)
    input = numpy.empty((1, ydim, zdim, 1, 1, 1), dtype=numpy.float64, order='F')
    expected = numpy.empty((1, ydim, zdim, 1, 1, 1), dtype=numpy.float64, order='F')
    sindex = 0
    iindex = 0
    for j in range(ydim):
        for k in range(zdim):
            if ((iindex % 13) == 3) or (sindex >= samplesize):
                input[0, j, k, 0, 0, 0] = inpbdfs[0]
                expected[0, j, k, 0, 0, 0] = resbdf
            else:
                input[0, j, k, 0, 0, 0] = sample[sindex]
                expected[0, j, k, 0, 0, 0] = zscores[sindex]
                sindex += 1
            iindex += 1
    if sindex != samplesize:
        raise ValueError("Unexpected final sindex of %d (ydim,zdim too small)" % sindex)
    result = -7777.0 * numpy.ones((1, ydim, zdim, 1, 1, 1), dtype=numpy.float64, order='F')

    # call ferret_compute and check the results
    ferret_compute(0, result, resbdf, (input, ), inpbdfs)
    if not numpy.allclose(result, expected, rtol=2.0E-7, atol=2.0E-7):
        print("expected (flattened) =\n%s" % str(expected.reshape(-1, order='F')))
        print("result (flattened) =\n%s" % str(result.reshape(-1, order='F')))
        raise ValueError("Unexpected result")

    # All successful
    print("Success")




'''
PipedImagerPQ is a graphics viewer application written in PyQt
that receives its images and commands primarily from another
application through a pipe.  A limited number of commands are
provided by the viewer itself to allow saving and some manipulation
of the displayed image.  The controlling application, however, may
be unaware of these modifications made to the image.

PipedImagerPQProcess is used to create and run a PipedImagerPQ.

This package was developed by the Thermal Modeling and Analysis
Project (TMAP) of the National Oceanographic and Atmospheric
Administration's (NOAA) Pacific Marine Environmental Lab (PMEL).
'''

from __future__ import print_function

import sys
import os
import time
import signal

# First try to import PySide2, then try PyQt5 if that fails, and finally try PyQt4 if that fails
try:
    import PySide2
    PYTHONQT_VERSION = 'PySide2'
except ImportError:
    try:
        import PyQt5
        PYTHONQT_VERSION = 'PyQt5'
    except ImportError:
        import PyQt4
        PYTHONQT_VERSION = 'PyQt4'

# Now that the Python Qt version is determined, import the parts
# allowing any import errors to propagate out
if PYTHONQT_VERSION == 'PySide2':
    from PySide2.QtCore    import Qt, QPointF, QRectF, QSize, QTimer
    from PySide2.QtGui     import QBrush, QColor, QImage, QPainter, \
                                  QPalette, QPen, QPixmap, QPolygonF
    from PySide2.QtWidgets import QAction, QApplication, QDialog, \
                                  QFileDialog, QLabel, QMainWindow, \
                                  QMessageBox, QPushButton, QScrollArea
elif PYTHONQT_VERSION == 'PyQt5':
    from PyQt5.QtCore    import Qt, QPointF, QRectF, QSize, QTimer
    from PyQt5.QtGui     import QBrush, QColor, QImage, QPainter, \
                                QPalette, QPen, QPixmap, QPolygonF
    from PyQt5.QtWidgets import QAction, QApplication, QDialog, \
                                QFileDialog, QLabel, QMainWindow, \
                                QMessageBox, QPushButton, QScrollArea
else:
    from PyQt4.QtCore import Qt, QPointF, QRectF, QSize, QTimer
    from PyQt4.QtGui  import QAction, QApplication, QBrush, QColor, QDialog, \
                             QFileDialog, QImage, QLabel, QMainWindow, \
                             QMessageBox, QPainter, QPalette, QPen, QPixmap, \
                             QPolygonF, QPushButton, QScrollArea

import multiprocessing

from pipedviewer import WINDOW_CLOSED_MESSAGE
from pipedviewer.cmndhelperpq import CmndHelperPQ
from pipedviewer.scaledialogpq import ScaleDialogPQ


class PipedImagerPQ(QMainWindow):
    '''
    A PyQt graphics viewer that receives images and commands through
    a pipe.

    A command is a dictionary with string keys.  For example,
        { "action":"save",
          "filename":"ferret.png",
          "fileformat":"png" }

    The command { "action":"exit" } will shutdown the viewer.
    '''

    def __init__(self, cmndpipe, rspdpipe):
        '''
        Create a PyQt viewer which reads commands from the Pipe
        cmndpipe and writes responses back to rspdpipe.
        '''
        super(PipedImagerPQ, self).__init__()
        self.__cmndpipe = cmndpipe
        self.__rspdpipe = rspdpipe
        # ignore Ctrl-C
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # unmodified image for creating the scene
        self.__sceneimage = None
        # bytearray of data for the above image
        self.__scenedata = None
        # flag set if in the process of reading image data from commands
        self.__loadingimage = False
        # width and height of the unmodified scene image
        # when the image is defined
        # initialize the width and height to values that will create
        # a viewer (mainWindow) of the right size
        self.__scenewidth = int(10.8 * self.physicalDpiX())
        self.__sceneheight = int(8.8 * self.physicalDpiY())
        # by default pay attention to any alpha channel values in colors
        self.__noalpha = False
        # initial default color for the background (opaque white)
        self.__lastclearcolor = QColor(0xFFFFFF)
        self.__lastclearcolor.setAlpha(0xFF)
        # scaling factor for creating the displayed scene
        self.__scalefactor = 1.0
        # automatically adjust the scaling factor to fit the window frame?
        self.__autoscale = True
        # minimum label width and height (for minimum scaling factor)
        # and minimum image width and height (for error checking)
        self.__minsize = 128
        # create the label, that will serve as the canvas, in a scrolled area
        self.__scrollarea = QScrollArea(self)
        self.__label = QLabel(self.__scrollarea)
        # set the initial label size and other values for the scrolled area
        self.__label.setMinimumSize(self.__scenewidth, self.__sceneheight)
        self.__label.resize(self.__scenewidth, self.__sceneheight)
        # setup the scrolled area
        self.__scrollarea.setWidget(self.__label)
        self.__scrollarea.setBackgroundRole(QPalette.Dark)
        self.setCentralWidget(self.__scrollarea)
        # default file name and format for saving the image
        self.__lastfilename = "ferret.png"
        self.__lastformat = "png"
        # command helper object
        self.__helper = CmndHelperPQ(self)
        # create the menubar
        self.__scaleact = QAction(self.tr("&Scale"), self,
                                shortcut=self.tr("Ctrl+S"),
                                statusTip=self.tr("Scale the image (canvas and image change size)"),
                                triggered=self.inquireSceneScale)
        self.__saveact = QAction(self.tr("Save &As..."), self,
                                shortcut=self.tr("Ctrl+A"),
                                statusTip=self.tr("Save the image to file"),
                                triggered=self.inquireSaveFilename)
        self.__redrawact = QAction(self.tr("&Redraw"), self,
                                shortcut=self.tr("Ctrl+R"),
                                statusTip=self.tr("Clear and redraw the image"),
                                triggered=self.redrawScene)
        self.__aboutact = QAction(self.tr("&About"), self,
                                statusTip=self.tr("Show information about this viewer"),
                                triggered=self.aboutMsg)
        self.__aboutqtact = QAction(self.tr("About &Qt"), self,
                                statusTip=self.tr("Show information about the Qt library"),
                                triggered=self.aboutQtMsg)
        self.createMenus()
        # set the initial size of the viewer
        self.__framedelta = 4
        mwwidth = self.__scenewidth + self.__framedelta
        mwheight = self.__sceneheight + self.__framedelta \
                 + self.menuBar().height() \
                 + self.statusBar().height()
        self.resize(mwwidth, mwheight)
        # check the command queue any time there are no window events to deal with
        self.__timer = QTimer(self)
        self.__timer.timeout.connect(self.checkCommandPipe)
        self.__timer.setInterval(0)
        self.__timer.start()

    def createMenus(self):
        '''
        Create the menu items for the viewer
        using the previously created actions.
        '''
        menuBar = self.menuBar()
        sceneMenu = menuBar.addMenu(menuBar.tr("&Image"))
        sceneMenu.addAction(self.__scaleact)
        sceneMenu.addAction(self.__saveact)
        sceneMenu.addAction(self.__redrawact)
        helpMenu = menuBar.addMenu(menuBar.tr("&Help"))
        helpMenu.addAction(self.__aboutact)
        helpMenu.addAction(self.__aboutqtact)

    def resizeEvent(self, event):
        '''
        Monitor resizing in case auto-scaling of the image is selected.
        '''
        if self.__autoscale:
            if self.autoScaleScene():
                # continue with the window resize
                event.accept()
            else:
                # another resize coming in, so ignore this one
                event.ignore()
        else:
            # continue with the window resize
            event.accept()

    def closeEvent(self, event):
        '''
        Clean up and send the WINDOW_CLOSED_MESSAGE on the response pipe
        before closing the window.
        '''
        self.__timer.stop()
        self.__cmndpipe.close()
        try:
            try:
                self.__rspdpipe.send(WINDOW_CLOSED_MESSAGE)
            finally:
                self.__rspdpipe.close()
        except Exception:
            pass
        event.accept()

    def exitViewer(self):
        '''
        Close and exit the viewer.
        '''
        self.close()

    def aboutMsg(self):
        QMessageBox.about(self, self.tr("About PipedImagerPQ"),
            self.tr("\n" \
            "PipedImagerPQ is a graphics viewer application that receives its " \
            "displayed image and commands primarily from another application " \
            "through a pipe.  A limited number of commands are provided by the " \
            "viewer itself to allow saving and some manipulation of the " \
            "displayed image.  The controlling application, however, may be " \
            "unaware of these modifications made to the image. " \
            "\n\n" \
            "PipedImagerPQ was developed by the Thermal Modeling and Analysis " \
            "Project (TMAP) of the National Oceanographic and Atmospheric " \
            "Administration's (NOAA) Pacific Marine Environmental Lab (PMEL). "))

    def aboutQtMsg(self):
        QMessageBox.aboutQt(self, self.tr("About Qt"))

    def ignoreAlpha(self):
        '''
        Return whether the alpha channel in colors should always be ignored.
        '''
        return self.__noalpha

    def updateScene(self):
        '''
        Clear the displayed scene using self.__lastclearcolor,
        then draw the scaled current image.
        '''
        # get the scaled scene size
        labelwidth = int(self.__scalefactor * self.__scenewidth + 0.5)
        labelheight = int(self.__scalefactor * self.__sceneheight + 0.5)
        # Create the new pixmap for the label to display
        newpixmap = QPixmap(labelwidth, labelheight)
        newpixmap.fill(self.__lastclearcolor)
        if self.__sceneimage != None:
            # Draw the scaled image to the pixmap
            mypainter = QPainter(newpixmap)
            trgrect = QRectF(0.0, 0.0, float(labelwidth),
                                       float(labelheight))
            srcrect = QRectF(0.0, 0.0, float(self.__scenewidth),
                                       float(self.__sceneheight))
            mypainter.drawImage(trgrect, self.__sceneimage, srcrect, Qt.AutoColor)
            mypainter.end()
        # Assign the new pixmap to the label
        self.__label.setPixmap(newpixmap)
        # set the label size and values
        # so the scrollarea knows of the new size
        self.__label.setMinimumSize(labelwidth, labelheight)
        self.__label.resize(labelwidth, labelheight)
        # update the label from the new pixmap
        self.__label.update()

    def clearScene(self, bkgcolor=None):
        '''
        Deletes the scene image and fills the label with bkgcolor.
        If bkgcolor is None or an invalid color, the color used is
        the one used from the last clearScene or redrawScene call
        with a valid color (or opaque white if a color has never
        been specified).
        '''
        # get the color to use for clearing (the background color)
        if bkgcolor:
            if bkgcolor.isValid():
                self.__lastclearcolor = bkgcolor
        # Remove the image and its bytearray
        self.__sceneimage = None
        self.__scenedata = None
        # Update the scene label using the current clearing color and image
        self.updateScene()

    def redrawScene(self, bkgcolor=None):
        '''
        Clear and redraw the displayed scene.
        '''
        # get the background color
        if bkgcolor:
            if bkgcolor.isValid():
                self.__lastclearcolor = bkgcolor
        # Update the scene label using the current clearing color and image
        QApplication.setOverrideCursor(Qt.WaitCursor)
        self.statusBar().showMessage( self.tr("Redrawing image") )
        try:
            self.updateScene()
        finally:
            self.statusBar().clearMessage()
            QApplication.restoreOverrideCursor()

    def resizeScene(self, width, height):
        '''
        Resize the scene to the given width and height in units of pixels.
        If the size changes, this deletes the current image and clear the
        displayed scene.
        '''
        newwidth = int(width + 0.5)
        if newwidth < self.__minsize:
            newwidth = self.__minsize
        newheight = int(height + 0.5)
        if newheight < self.__minsize:
            newheight = self.__minsize
        if (newwidth != self.__scenewidth) or (newheight != self.__sceneheight):
            # set the new size for the empty scene
            self.__scenewidth = newwidth
            self.__sceneheight = newheight
            # If auto-scaling, set scaling factor to 1.0 and resize the window
            if self.__autoscale:
                self.__scalefactor = 1.0
                barheights = self.menuBar().height() + self.statusBar().height()
                self.resize(newwidth + self.__framedelta,
                            newheight + self.__framedelta + barheights)
            # clear the scene with the last clearing color
            self.clearScene(None)

    def loadNewSceneImage(self, imageinfo):
        '''
        Create a new scene image from the information given in this
        and subsequent dictionaries imageinfo.  The image is created
        from multiple calls to this function since there is a limit
        on the size of a single object passed through a pipe.

        The first imageinfo dictionary given when creating an image
        must define the following key and value pairs:
            "width": width of the image in pixels
            "height": height of the image in pixels
            "stride": number of bytes in one line of the image
                      in the bytearray
        The scene image data is initialized to all zero (transparent)
        at this time.

        This initialization call must be followed by (multiple) calls
        to this method with imageinfo dictionaries defining the key
        and value pairs:
            "blocknum": data block number (1, 2, ... numblocks)
            "numblocks": total number of image data blocks
            "startindex": index in the bytearray of image data
                          where this block of image data starts
            "blockdata": this block of data as a bytearray

        On receipt of the last block of data (blocknum == numblocks)
        the scene image will be created and the scene will be updated.

        Raises:
            KeyError - if one of the above keys is not given
            ValueError - if a value for a key is not valid
        '''
        if not self.__loadingimage:
            # prepare for a new image data from subsequent calls
            # get dimensions of the new image
            myimgwidth = int( imageinfo["width"] )
            myimgheight = int( imageinfo["height"] )
            myimgstride = int( imageinfo["stride"] )
            if (myimgwidth < self.__minsize) or (myimgheight < self.__minsize):
                raise ValueError("image width and height cannot be less than %s" % str(self.__minsize))
            # Newer PyQt versions allow separate specification of the stride
            if myimgstride != 4 * myimgwidth:
                raise ValueError("image stride is not four times the image width")
            # create the bytearray to contain the new scene data
            # automatically initialized to zero
            self.__scenedata = bytearray(myimgstride * myimgheight)
            self.__scenewidth = myimgwidth
            self.__sceneheight = myimgheight
            # set the flag for subsequent calls to this method
            self.__loadingimage = True
            # change the cursor to warn the user this may take some time
            QApplication.setOverrideCursor(Qt.WaitCursor)
            # put up an appropriate status message
            self.statusBar().showMessage( self.tr("Loading new image") )
            return
        # loading an image; add the next block of data
        myblocknum = int( imageinfo["blocknum"] )
        mynumblocks = int( imageinfo["numblocks"] )
        mystartindex = int( imageinfo["startindex"] )
        myblockdata = imageinfo["blockdata"]
        if (myblocknum < 1) or (myblocknum > mynumblocks):
            self.statusBar().clearMessage()
            QApplication.restoreOverrideCursor()
            raise ValueError("invalid image data block number or number of blocks")
        if (mystartindex < 0) or (mystartindex >= len(self.__scenedata)):
            self.statusBar().clearMessage()
            QApplication.restoreOverrideCursor()
            raise ValueError("invalid start index for an image data block")
        myblocksize = len(myblockdata)
        myendindex = mystartindex + myblocksize
        if (myblocksize < 1) or (myendindex > len(self.__scenedata)):
            self.statusBar().clearMessage()
            QApplication.restoreOverrideCursor()
            raise ValueError("invalid length of an image data block")
        # update the status message to show progress
        self.statusBar().showMessage( self.tr("Loading new image (block %s of %s)" % \
                                              (str(myblocknum),str(mynumblocks))) )
        # assign the data
        self.__scenedata[mystartindex:myendindex] = myblockdata
        # if this is the last block of data, create and display the scene image
        if myblocknum == mynumblocks:
            self.__loadingimage = False
            self.statusBar().showMessage( self.tr("Creating new image") )
            try:
                self.__sceneimage = QImage(self.__scenedata,
                                           self.__scenewidth,
                                           self.__sceneheight,
                                           QImage.Format_ARGB32_Premultiplied)
                self.statusBar().showMessage( self.tr("Drawing new image") )
                # update the displayed scene in the label
                self.updateScene()
            finally:
                # clear the status message
                self.statusBar().clearMessage()
                # restore the cursor back to normal
                QApplication.restoreOverrideCursor()

    def inquireSceneScale(self):
        '''
        Prompt the user for the desired scaling factor for the scene.
        '''
        labelwidth = int(self.__scenewidth * self.__scalefactor + 0.5)
        labelheight = int(self.__sceneheight * self.__scalefactor + 0.5)
        scaledlg = ScaleDialogPQ(self.__scalefactor, labelwidth, labelheight,
                        self.__minsize, self.__minsize, self.__autoscale, self)
        if scaledlg.exec_():
            (newscale, autoscale, okay) = scaledlg.getValues()
            if okay:
                if autoscale:
                    self.__autoscale = True
                    self.autoScaleScene()
                else:
                    self.__autoscale = False
                    self.scaleScene(newscale, False)

    def autoScaleScene(self):
        '''
        Selects a scaling factor that maximizes the scene within the window
        frame without requiring scroll bars.  Intended to be called when
        the window size is changed by the user and auto-scaling is turn on.

        Returns:
            True if scaling of this scene is done (no window resize)
            False if the a window resize command was issued
        '''
        barheights = self.menuBar().height() + self.statusBar().height()

        # get the size for the central widget
        cwheight = self.height() - barheights - self.__framedelta
        heightsf = float(cwheight) / float(self.__sceneheight)

        cwwidth = self.width() - self.__framedelta
        widthsf = float(cwwidth) / float(self.__scenewidth)

        if heightsf < widthsf:
            factor = heightsf
        else:
            factor = widthsf

        newcwheight = int(factor * self.__sceneheight + 0.5)
        newcwwidth = int(factor * self.__scenewidth + 0.5)

        # if the window does not have the correct aspect ratio, resize it so
        # it will; this will generate another call to this method.  Otherwise,
        # scale the scene and be done.
        if self.isMaximized() or \
           ( (abs(cwheight - newcwheight) <= self.__framedelta) and \
             (abs(cwwidth - newcwwidth) <= self.__framedelta) ):
            self.scaleScene(factor, False)
            return True
        else:
            self.resize(newcwwidth + self.__framedelta,
                        newcwheight + self.__framedelta + barheights)
            return False

    def scaleScene(self, factor, resizewin):
        '''
        Scales both the horizontal and vertical directions by factor.
        Scaling factors are not accumulative.  So if the scene was
        already scaled, that scaling is "removed" before this scaling
        factor is applied.  If resizewin is True, the main window is
        resized to accommodate this new scaled scene size.

        If factor is zero, just switch to auto-scaling at the current
        window size.  If factor is negative, rescale using the absolute
        value (possibly resizing the window) then switch to auto-scaling.
        '''
        fltfactor = float(factor)
        if fltfactor != 0.0:
            if resizewin:
                # from command - turn off autoscaling for the following
                # then turn back on if appropriate
                self.__autoscale = False
            newfactor = abs(fltfactor)
            newlabwidth = int(newfactor * self.__scenewidth + 0.5)
            newlabheight = int(newfactor * self.__sceneheight + 0.5)
            if (newlabwidth < self.__minsize) or (newlabheight < self.__minsize):
                # Set to minimum size
                if self.__scenewidth <= self.__sceneheight:
                    newfactor = float(self.__minsize) / float(self.__scenewidth)
                else:
                    newfactor = float(self.__minsize) / float(self.__sceneheight)
                newlabwidth = int(newfactor * self.__scenewidth + 0.5)
                newlabheight = int(newfactor * self.__sceneheight + 0.5)
            oldlabwidth = int(self.__scalefactor * self.__scenewidth + 0.5)
            oldlabheight = int(self.__scalefactor * self.__sceneheight + 0.5)
            if (newlabwidth != oldlabwidth) or (newlabheight != oldlabheight):
                # Set the new scaling factor
                self.__scalefactor = newfactor
                # Update the scene label using the current clearing color and image
                QApplication.setOverrideCursor(Qt.WaitCursor)
                self.statusBar().showMessage( self.tr("Scaling image") )
                try:
                    self.updateScene()
                finally:
                    self.statusBar().clearMessage()
                    QApplication.restoreOverrideCursor()
            if resizewin:
                # resize the main window (if possible)
                barheights = self.menuBar().height() + self.statusBar().height()
                mwheight = newlabheight + barheights + self.__framedelta
                mwwidth = newlabwidth + self.__framedelta
                # Do not exceed the available real estate on the screen.
                # If autoscaling is in effect, the resize will trigger
                # any required adjustments.
                scrnrect = QApplication.desktop().availableGeometry()
                if mwwidth > 0.95 * scrnrect.width():
                    mwwidth = int(0.9 * scrnrect.width() + 0.5)
                if mwheight > 0.95 * scrnrect.height():
                    mwheight = int(0.9 * scrnrect.height() + 0.5)
                self.resize(mwwidth, mwheight)
        if fltfactor <= 0.0:
            # From command - turn on autoscaling
            self.__autoscale = True
            self.autoScaleScene();

    def inquireSaveFilename(self):
        '''
        Prompt the user for the name of the file into which to save the scene.
        The file format will be determined from the filename extension.
        '''
        formattypes = [ ( "png",
                          "PNG - Portable Networks Graphics (*.png)" ),
                        ( "jpeg",
                          "JPEG - Joint Photographic Experts Group (*.jpeg *.jpg *.jpe)" ),
                        ( "tiff",
                          "TIFF - Tagged Image File Format (*.tiff *.tif)" ),
                        ( "bmp",
                          "BMP - Windows Bitmap (*.bmp)" ),
                        ( "ppm",
                          "PPM - Portable Pixmap (*.ppm)" ),
                        ( "xpm",
                          "XPM - X11 Pixmap (*.xpm)" ),
                        ( "xbm",
                          "XBM - X11 Bitmap (*.xbm)" ), ]
        filters = ";;".join( [ t[1] for t in formattypes ] )
        if PYTHONQT_VERSION == 'PyQt4':
            # getSaveFileNameAndFilter; tr returns QStrings in PyQt4 (Python2)
            (fileName, fileFilter) = QFileDialog.getSaveFileNameAndFilter(self,
                 self.tr("Save the current image as "), self.tr(self.__lastfilename), self.tr(filters))
        else:
            # getSaveFileName; tr returns Python unicode strings in PySide2 and PyQt5 (Python3)
            (fileName, fileFilter) = QFileDialog.getSaveFileName(self,
                self.tr("Save the current image as "), self.tr(self.__lastfilename), self.tr(filters))
        if fileName:
            for (fmt, fmtQName) in formattypes:
                if self.tr(fmtQName) == fileFilter:
                    fileFormat = fmt
                    break
            else:
                raise RuntimeError("Unexpected file format name '%s'" % fileFilter)
            self.saveSceneToFile(fileName, fileFormat, None, None)
            self.__lastfilename = fileName
            self.__lastformat = fileFormat

    def saveSceneToFile(self, filename, imageformat, transparent, rastsize):
        '''
        Save the current scene to the named file.

        If imageformat is empty or None, the format is guessed from
        the filename extension.

        If transparent is False, the entire scene is initialized
        to the last clearing color.

        If given, rastsize is the pixels size of the saved image.
        If rastsize is not given, the saved image will be saved
        at the current scaled image size.
        '''
        # This could be called when there is no image present.
        # If this is the case, ignore the call.
        if ( self.__sceneimage == None ):
            return
        if not imageformat:
            # Guess the image format from the filename extension
            # This is only done to silently change gif to png
            fileext = ( os.path.splitext(filename)[1] ).lower()
            if fileext == '.gif':
                myformat = 'gif'
            else:
                # let QImage figure out the format
                myformat = None
        else:
            myformat = imageformat.lower()

        if myformat == 'gif':
            # Silently convert gif filename and format to png
            myformat = 'png'
            myfilename = os.path.splitext(filename)[0] + ".png"
        else:
            myfilename = filename
        # set the cursor and status message to indicate a save is happending
        QApplication.setOverrideCursor(Qt.WaitCursor)
        self.statusBar().showMessage( self.tr("Saving image") )
        try:
            if rastsize:
                imagewidth = int(rastsize.width() + 0.5)
                imageheight = int(rastsize.height() + 0.5)
            else:
                imagewidth = int(self.__scenewidth * self.__scalefactor + 0.5)
                imageheight = int(self.__sceneheight * self.__scalefactor + 0.5)
            myimage = QImage( QSize(imagewidth, imageheight),
                            QImage.Format_ARGB32_Premultiplied )
            # Initialize the image
            if not transparent:
                # Clear the image with self.__lastclearcolor
                fillint = self.__helper.computeARGB32PreMultInt(self.__lastclearcolor)
            else:
                fillint = 0
            myimage.fill(fillint)
            # draw the scaled scene to this QImage
            mypainter = QPainter(myimage)
            trgrect = QRectF(0.0, 0.0, float(imagewidth),
                                       float(imageheight))
            srcrect = QRectF(0.0, 0.0, float(self.__scenewidth),
                                       float(self.__sceneheight))
            mypainter.drawImage(trgrect, self.__sceneimage, srcrect, Qt.AutoColor)
            mypainter.end()
            # save the image to file
            if not myimage.save(myfilename, myformat):
                raise ValueError("Unable to save the plot as " + myfilename)
        finally:
            self.statusBar().clearMessage()
            QApplication.restoreOverrideCursor()

    def checkCommandPipe(self):
        '''
        Get and perform commands waiting in the pipe.
        Stop when no more commands or if more than 50
        milliseconds have passed.
        '''
        try:
            if (sys.version_info[0] >= 3) and (sys.version_info[1] >= 3):
                starttime = time.process_time()
            else:
                starttime = time.clock()
            # Wait up to 2 milliseconds waiting for a command.
            # This prevents unchecked spinning when there is
            # nothing to do (Qt immediately calling this method
            # again only for this method to immediately return).
            while self.__cmndpipe.poll(0.002):
                cmnd = self.__cmndpipe.recv()
                self.processCommand(cmnd)
                # Continue to try to process commands until
                # more than 50 milliseconds have passed.
                # This reduces Qt overhead when there are lots
                # of commands waiting in the queue.
                if (sys.version_info[0] >= 3) and (sys.version_info[1] >= 3):
                    elapsed = time.process_time() - starttime
                else:
                    elapsed = time.clock() - starttime
                if elapsed > 0.050:
                    break
        except EOFError:
            # Assume PyFerret has shut down
            self.exitViewer()
        except Exception:
            # Some problem, but presumably still functional
            (exctype, excval) = sys.exc_info()[:2]
            try:
                if excval:
                    self.__rspdpipe.send("**ERROR %s: %s" % (str(exctype), str(excval)))
                else:
                    self.__rspdpipe.send("**ERROR %s" % str(exctype))
            except Exception:
                pass

    def processCommand(self, cmnd):
        '''
        Examine the action of cmnd and call the appropriate
        method to deal with this command.  Raises a KeyError
        if the "action" key is missing.
        '''
        try:
            cmndact = cmnd["action"]
        except KeyError:
            raise ValueError("Unknown command '%s'" % str(cmnd))

        if cmndact == "clear":
            try:
                bkgcolor = self.__helper.getColorFromCmnd(cmnd)
            except KeyError:
                bkgcolor = None
            self.clearScene(bkgcolor)
        elif cmndact == "exit":
            self.exitViewer()
        elif cmndact == "hide":
            self.showMinimized()
        elif cmndact == "screenInfo":
            scrnrect = QApplication.desktop().availableGeometry()
            info = ( self.physicalDpiX(), self.physicalDpiY(),
                     scrnrect.width(), scrnrect.height() )
            self.__rspdpipe.send(info)
        elif cmndact == "redraw":
            try:
                bkgcolor = self.__helper.getColorFromCmnd(cmnd)
            except KeyError:
                bkgcolor = None
            self.redrawScene(bkgcolor)
        elif cmndact == "rescale":
            self.scaleScene(float(cmnd["factor"]), True)
        elif cmndact == "resize":
            mysize = self.__helper.getSizeFromCmnd(cmnd)
            self.resizeScene(mysize.width(), mysize.height())
        elif cmndact == "newImage":
            self.loadNewSceneImage(cmnd)
        elif cmndact == "save":
            filename = cmnd["filename"]
            fileformat = cmnd.get("fileformat", None)
            try:
                bkgcolor = self.__helper.getColorFromCmnd(cmnd)
            except KeyError:
                bkgcolor = None
            rastsize = self.__helper.getSizeFromCmnd(cmnd["rastsize"])
            self.saveSceneToFile(filename, fileformat, bkgcolor, rastsize)
        elif cmndact == "setTitle":
            self.setWindowTitle(cmnd["title"])
        elif cmndact == "imgname":
            myvalue = cmnd.get("name", None)
            if myvalue:
                self.__lastfilename = myvalue
            myvalue = cmnd.get("format", None)
            if myvalue:
                self.__lastformat = myvalue.lower()
        elif cmndact == "show":
            if not self.isVisible():
                self.show()
        elif cmndact == "noalpha":
            # ignore any alpha channel values in colors
            self.__noalpha = True
        else:
            raise ValueError("Unknown command action %s" % str(cmndact))


class PipedImagerPQProcess(multiprocessing.Process):
    '''
    A Process specifically tailored for creating a PipedImagerPQ.
    '''
    def __init__(self, cmndpipe, rspdpipe):
        '''
        Create a Process that will produce a PipedImagerPQ
        attached to the given Pipes when run.
        '''
        super(PipedImagerPQProcess,self).__init__(group=None, target=None, name='PipedImagerPQ')
        self.__cmndpipe = cmndpipe
        self.__rspdpipe = rspdpipe
        self.__app = None
        self.__viewer = None

    def run(self):
        '''
        Create a PipedImagerPQ that is attached
        to the Pipe of this instance.
        '''
        self.__app = QApplication(["PipedImagerPQ"])
        self.__viewer = PipedImagerPQ(self.__cmndpipe, self.__rspdpipe)
        myresult = self.__app.exec_()
        sys.exit(myresult)


#
# The following are for testing this module
#

class _CommandSubmitterPQ(QDialog):
    '''
    Testing dialog for controlling the addition of commands to a pipe.
    Used for testing PipedImagerPQ in the same process as the viewer.
    '''
    def __init__(self, parent, cmndpipe, rspdpipe, cmndlist):
        '''
        Create a QDialog with a single QPushButton for controlling
        the submission of commands from cmndlist to cmndpipe.
        '''
        super(_CommandSubmitterPQ,self).__init__(parent)
        self.__cmndlist = cmndlist
        self.__cmndpipe = cmndpipe
        self.__rspdpipe = rspdpipe
        self.__nextcmnd = 0
        self.__button = QPushButton("Submit next command", self)
        self.__button.pressed.connect(self.submitNextCommand)
        self.show()

    def submitNextCommand(self):
        '''
        Submit the next command from the command list to the command pipe,
        or shutdown if there are no more commands to submit.
        '''
        try:
            cmndstr = str(self.__cmndlist[self.__nextcmnd])
            if len(cmndstr) > 188:
                cmndstr = cmndstr[:188] + '...'
            print("Command: %s" % cmndstr)
            self.__cmndpipe.send(self.__cmndlist[self.__nextcmnd])
            self.__nextcmnd += 1
            while self.__rspdpipe.poll(0.1):
                print("Response: %s" % str(self.__rspdpipe.recv()))
        except IndexError:
            self.__rspdpipe.close()
            self.__cmndpipe.close()
            self.close()


def _test_pipedimagerpq():
    # vertices of a pentagon (roughly) centered in a 1000 x 1000 square
    pentagonpts = ( (504.5, 100.0), (100.0, 393.9),
                    (254.5, 869.4), (754.5, 869.4),
                    (909.0, 393.9),  )
    linepts = ( (350,  50),
                (200, 150),
                (400, 250),
                (300, 350),
                (150, 250),
                (100, 450) )
    # start PyQt
    testapp = QApplication(["PipedImagerPQ"])
    # create the list of commands to submit
    drawcmnds = []
    drawcmnds.append( { "action":"setTitle", "title":"Tester" } )
    drawcmnds.append( { "action":"show" } )
    drawcmnds.append( { "action":"clear", "color":"black"} )
    drawcmnds.append( { "action":"screenInfo"} )
    # create the image to be displayed
    testimage = QImage(500, 500, QImage.Format_ARGB32_Premultiplied)
    # initialize a black background
    testimage.fill(0xFF000000)
    # draw some things in the image
    testpainter = QPainter(testimage)
    testpainter.setBrush( QBrush(QColor(0, 255, 0, 128), Qt.SolidPattern) )
    testpainter.setPen( QPen(QBrush(QColor(255, 0, 0, 255), Qt.SolidPattern),
                         5.0, Qt.SolidLine, Qt.SquareCap, Qt.MiterJoin) )
    testpainter.drawRect( QRectF(5.0, 255.0, 240.0, 240.0) )
    testpainter.setBrush( QBrush(QColor(0, 0, 255, 255), Qt.SolidPattern) )
    testpainter.setPen( QPen(QBrush(QColor(0, 0, 0, 255), Qt.SolidPattern),
                         5.0, Qt.DashLine, Qt.RoundCap, Qt.RoundJoin) )
    testpainter.drawPolygon( QPolygonF(
            [ QPointF(.25 * ptx, .25 * pty + 250) for (ptx, pty) in pentagonpts ] ) )
    testpainter.setBrush( Qt.NoBrush )
    testpainter.setPen( QPen(QBrush(QColor(255, 255, 255, 255), Qt.SolidPattern),
                         3.0, Qt.DashLine, Qt.RoundCap, Qt.RoundJoin) )
    testpainter.drawPolyline( QPolygonF(
            [ QPointF(pts, pty) for (pts, pty) in linepts ] ) )
    testpainter.end()
    # add the image command
    testimgwidth = testimage.width()
    testimgheight = testimage.height()
    testimgstride = testimage.bytesPerLine()
    # not a good way to get the pixel data
    testimgdata = bytearray(testimgheight * testimgstride)
    k = 0
    for pty in range(testimgheight):
        for ptx in range(testimgwidth):
            pixval = testimage.pixel(ptx, pty)
            (aval, rgbval) = divmod(pixval, 256 * 256 * 256)
            (rval, gbval) = divmod(rgbval, 256 * 256)
            (gval, bval) = divmod(gbval, 256)
            testimgdata[k] = bval
            k += 1
            testimgdata[k] = gval
            k += 1
            testimgdata[k] = rval
            k += 1
            testimgdata[k] = aval
            k += 1
    testblocksize = 4000
    testnumblocks = (testimgheight * testimgstride + testblocksize - 1) // testblocksize
    drawcmnds.append( { "action":"newImage",
                        "width":testimgwidth,
                        "height":testimgheight,
                        "stride":testimgstride } )
    for k in range(testnumblocks):
        if k < (testnumblocks - 1):
            blkdata = testimgdata[k*testblocksize:(k+1)*testblocksize]
        else:
            blkdata = testimgdata[k*testblocksize:]
        drawcmnds.append( { "action":"newImage",
                            "blocknum":k+1,
                            "numblocks":testnumblocks,
                            "startindex":k*testblocksize,
                            "blockdata":blkdata } )
    # finish the command list
    drawcmnds.append( { "action":"show" } )
    drawcmnds.append( { "action":"exit" } )
    # create a PipedImagerPQ in this process
    (cmndrecvpipe, cmndsendpipe) = multiprocessing.Pipe(False)
    (rspdrecvpipe, rspdsendpipe) = multiprocessing.Pipe(False)
    testviewer = PipedImagerPQ(cmndrecvpipe, rspdsendpipe)
    # create a command submitter dialog
    tester = _CommandSubmitterPQ(testviewer, cmndsendpipe,
                                   rspdrecvpipe, drawcmnds)
    tester.show()
    # let it all run
    testresult = testapp.exec_()
    if testresult != 0:
        sys.exit(testresult)

if __name__ == "__main__":
    _test_pipedimagerpq()



"""
Returns the array of cumulative density function values
for the Gamma probability distribution
using the given arrays for the abscissa or template
values and each of the parameters values.
"""
import numpy
import pyferret
import pyferret.stats

DISTRIB_NAME = "Gamma"
FUNC_NAME = "cdf"


def ferret_init(id):
    """
    Initialization for the stats_gamma_cdf Ferret PyEF
    """
    return pyferret.stats.getinitdict(DISTRIB_NAME, FUNC_NAME)


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Result array assignment for the stats_gamma_cdf Ferret PyEF
    """
    pyferret.stats.assignresultsarray(DISTRIB_NAME, FUNC_NAME,
                                      result, resbdf, inputs, inpbdfs)




"""
Creates a shapefile with a given root name using data from given
X, Y, and value arrays (curvilinear-type data).  The shapes are
quadrilaterals in the X,Y-plane derived from the X and Y arrays;
i.e., vertices are the (i,j), (i,j+1), (i+1,j+1), and (i+1,j)
elements of the X and Y coordinates.  The value associated with
each quadrilateral comes from the value array; i.e., the (i,j)
element of the value for the previously mentioned quadrilateral.
Quadrilaterals associated with missing values are omitted from
the shapefile.
"""

from __future__ import print_function

import shapefile
import pyferret
import pyferret.fershp

def ferret_init(efid):
    """
    Initialization for the shapefile_writexyval PyEF
    """
    retdict = { "numargs": 6,
                "descript": "Writes a shapefile of XY quadrilaterals from the curvilinear data arrays.",
                "restype": pyferret.FLOAT_ARRAY,
                "axes": ( pyferret.AXIS_ABSTRACT,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST, ),
                "argnames": ( "SHAPEFILE", "GRIDX", "GRIDY", "VALUE", "VALNAME", "MAPPRJ"),
                "argdescripts": ( "Shapefile name (any extension given is ignored)",
                                  "X values (longitudes) for the quad. grid; must be 2D on X and Y axes",
                                  "Y values (latitudes) for the quad. grid; must be 2D on X and Y axes",
                                  "Shape values; must be 2D on X and Y axes",
                                  "Name for the shape value",
                                  "Common name or WKT description of map projection; " \
                                      "if blank, WGS 84 is used", ),
                "argtypes": ( pyferret.STRING_ONEVAL,
                              pyferret.FLOAT_ARRAY,
                              pyferret.FLOAT_ARRAY,
                              pyferret.FLOAT_ARRAY,
                              pyferret.STRING_ONEVAL,
                              pyferret.STRING_ONEVAL, ),
                "influences": ( (False, False, False, False, False, False),
                                (False, False, False, False, False, False),
                                (False, False, False, False, False, False),
                                (False, False, False, False, False, False),
                                (False, False, False, False, False, False),
                                (False, False, False, False, False, False), ),
              }
    return retdict


def ferret_result_limits(efid):
    """
    Abstract axis limits for the shapefile_writexyval PyEF
    """
    return ( (1, 1), None, None, None, None, None, )


def ferret_compute(efid, result, resbdf, inputs, inpbdfs):
    """
    Create the shapefile named in inputs[0] using the grid X coordinates given
    in inputs[1], grid Y coordinates given in inputs[2], and shape values given
    in inputs[3].  The X,Y coordinates are used for the quadrilaterals vertices
    and must have an additional value along each dimension.  The value [i,j]
    is used for the quadrilateral with diagonal corners [i, j] and [i+1, j+1].
    Quadrilateral associated with missing values are omitted from the shapefile.
    The field name for the value in the shapefile given in inputs[4].  Either a
    common name or a WKT description of the map projection for the coordinates
    should be given in inputs[5].  If blank, WGS 84 is used.  If successful,
    fills result (which might as well be a 1x1x1x1 array) with zeros.  If a
    problem occurs, an error will be raised.
    """
    shapefile_name = inputs[0]
    grid_xs = inputs[1]
    grid_ys = inputs[2]
    grid_vals = inputs[3]
    missing_val = inpbdfs[3]
    field_name = inputs[4].strip()
    if not field_name:
        field_name = "VALUE"
    map_projection = inputs[5]

    # Verify the shapes are as expected
    if (grid_vals.shape[2] != 1) or (grid_vals.shape[3] != 1) or \
       (grid_vals.shape[4] != 1) or (grid_vals.shape[5] != 1):
        raise ValueError("The Z, T, E, and F axes of VALUE must be undefined or singleton axes")
    exp_shape = ( grid_vals.shape[0] + 1, grid_vals.shape[1] + 1, 1, 1, 1, 1 )
    if (grid_xs.shape != exp_shape) or (grid_ys.shape != exp_shape):
         raise ValueError('GRIDX and GRIDY must have one more value along both X and Y axes compared to VALUE')

    # Create polygons with a single field value
    sfwriter = shapefile.Writer(shapefile.POLYGON)
    sfwriter.field(field_name, "N", 20, 7)

    # Add the shapes with their values
    shape_written = False
    for j in range(grid_vals.shape[1]):
        for i in range(grid_vals.shape[0]):
            if grid_vals[i, j, 0, 0, 0, 0] != missing_val:
                shape_written = True
                pyferret.fershp.addquadxyvalues(sfwriter,
                         (grid_xs[i,   j,   0, 0, 0, 0], grid_ys[i,   j,   0, 0, 0, 0]),
                         (grid_xs[i,   j+1, 0, 0, 0, 0], grid_ys[i,   j+1, 0, 0, 0, 0]),
                         (grid_xs[i+1, j+1, 0, 0, 0, 0], grid_ys[i+1, j+1, 0, 0, 0, 0]),
                         (grid_xs[i+1, j,   0, 0, 0, 0], grid_ys[i+1, j,   0, 0, 0, 0]),
                         None, [ float(grid_vals[i, j, 0, 0, 0, 0]) ])
    if not shape_written:
        raise ValueError("All values are missing values")
    sfwriter.save(shapefile_name)

    # Create the .prj file from the map projection common name or the WKT description
    pyferret.fershp.createprjfile(map_projection, shapefile_name)
    result[:, :, :, :, :, :] = 0


#
# The following is only for testing this module from the command line
#
if __name__ == "__main__":
    import numpy
    import os

    shapefilename = "tripolar"
    fieldname = "AREA"
    wgs84_descript = 'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]]'

    # Real world longitudes and latitudes of tripolar coordinates X=80W:60E:10 + 100E:120W:10,Y=45N:90N:5
    geolon_c = numpy.array([
        [ -100.0,-100.0,-100.0,-100.0,-100.0,-100.0,-100.0,-100.0,  80.0,  80.0,
            80.0,  80.0,  80.0,  80.0,  80.0,  80.0,  80.0,  80.0,  80.0,  80.0,
            80.0,  80.0, 260.0, 260.0, 260.0, 260.0, 260.0, 260.0, 260.0, 260.0, ],
        [  -92.1, -87.7, -82.4, -75.7, -66.7, -53.8, -34.9, -10.0,  14.9,  33.8,
            46.7,  55.7,  62.4,  67.7,  72.1,  87.9,  92.3,  97.6, 104.3, 113.3,
           126.2, 145.1, 170.0, 194.9, 213.8, 226.7, 235.7, 242.4, 247.7, 252.1, ],
        [  -86.0, -78.5, -70.2, -60.9, -50.2, -38.1, -24.5, -10.0,   4.5,  18.1,
            30.2,  40.9,  50.2,  58.5,  66.0,  94.0, 101.5, 109.8, 119.1, 129.8,
           141.9, 155.5, 170.0, 184.5, 198.1, 210.2, 220.9, 230.2, 238.5, 246.0, ],
        [  -82.3, -73.1, -63.6, -53.7, -43.3, -32.5, -21.4, -10.0,   1.4,  12.5,
            23.3,  33.7,  43.6,  53.1,  62.3,  97.7, 106.9, 116.4, 126.3, 136.7,
           147.5, 158.6, 170.0, 181.4, 192.5, 203.3, 213.7, 223.6, 233.1, 242.3, ],
        [  -80.5, -70.6, -60.7, -50.7, -40.6, -30.5, -20.3, -10.0,   0.3,  10.5,
            20.6,  30.7,  40.7,  50.6,  60.5,  99.5, 109.4, 119.3, 129.3, 139.4,
           149.5, 159.7, 170.0, 180.3, 190.5, 200.6, 210.7, 220.7, 230.6, 240.5, ],
        [  -80.0, -70.0, -60.0, -50.0, -40.0, -30.0, -20.0, -10.0,   0.0,  10.0,
            20.0,  30.0,  40.0,  50.0,  60.0, 100.0, 110.0, 120.0, 130.0, 140.0,
           150.0, 160.0, 170.0, 180.0, 190.0, 200.0, 210.0, 220.0, 230.0, 240.0, ],
        [  -80.0, -70.0, -60.0, -50.0, -40.0, -30.0, -20.0, -10.0,   0.0,  10.0,
            20.0,  30.0,  40.0,  50.0,  60.0, 100.0, 110.0, 120.0, 130.0, 140.0,
           150.0, 160.0, 170.0, 180.0, 190.0, 200.0, 210.0, 220.0, 230.0, 240.0, ],
        [  -80.0, -70.0, -60.0, -50.0, -40.0, -30.0, -20.0, -10.0,   0.0,  10.0,
            20.0,  30.0,  40.0,  50.0,  60.0,  100.0, 110.0, 120.0, 130.0, 140.0,
           150.0, 160.0, 170.0, 180.0, 190.0, 200.0, 210.0, 220.0, 230.0, 240.0, ],
        [  -80.0, -70.0, -60.0, -50.0, -40.0, -30.0, -20.0, -10.0,   0.0,  10.0,
            20.0,  30.0,  40.0,  50.0,  60.0, 100.0, 110.0, 120.0, 130.0, 140.0,
           150.0, 160.0, 170.0, 180.0, 190.0, 200.0, 210.0, 220.0, 230.0, 240.0, ],
        [  -80.0, -70.0, -60.0, -50.0, -40.0, -30.0, -20.0, -10.0,   0.0,  10.0,
            20.0,  30.0,  40.0,  50.0,  60.0, 100.0, 110.0, 120.0, 130.0, 140.0,
           150.0, 160.0, 170.0, 180.0, 190.0, 200.0, 210.0, 220.0, 230.0, 240.0, ],
    ], dtype=numpy.float64)
    geolon_c = geolon_c.T[:, :, numpy.newaxis, numpy.newaxis, numpy.newaxis, numpy.newaxis]
    geolat_c = numpy.array([
        [ 72.35, 75.41, 78.20, 80.77, 83.20, 85.52, 87.78, 90.00, 87.78, 85.52,
          83.20, 80.77, 78.20, 75.41, 72.35, 72.35, 75.41, 78.20, 80.77, 83.20,
          85.52, 87.78, 90.00, 87.78, 85.52, 83.20, 80.77, 78.20, 75.41, 72.35, ],
        [ 71.85, 74.69, 77.25, 79.54, 81.58, 83.30, 84.53, 85.00, 84.53, 83.30,
          81.58, 79.54, 77.25, 74.69, 71.85, 71.85, 74.69, 77.25, 79.54, 81.58,
          83.30, 84.53, 85.00, 84.53, 83.30, 81.58, 79.54, 77.25, 74.69, 71.85, ],
        [ 70.51, 72.81, 74.83, 76.56, 77.99, 79.08, 79.76, 80.00, 79.76, 79.08,
          77.99, 76.56, 74.83, 72.81, 70.51, 70.51, 72.81, 74.83, 76.56, 77.99,
          79.08, 79.76, 80.00, 79.76, 79.08, 77.99, 76.56, 74.83, 72.81, 70.51, ],
        [ 68.71, 70.29, 71.67, 72.83, 73.76, 74.44, 74.86, 75.00, 74.86, 74.44,
          73.76, 72.83, 71.67, 70.29, 68.71, 68.71, 70.29, 71.67, 72.83, 73.76,
          74.44, 74.86, 75.00, 74.86, 74.44, 73.76, 72.83, 71.67, 70.29, 68.71, ],
        [ 66.80, 67.60, 68.30, 68.90, 69.37, 69.72, 69.93, 70.00, 69.93, 69.72,
          69.37, 68.90, 68.30, 67.60, 66.80, 66.80, 67.60, 68.30, 68.90, 69.37,
          69.72, 69.93, 70.00, 69.93, 69.72, 69.37, 68.90, 68.30, 67.60, 66.80, ],
        [ 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00,
          65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00,
          65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, 65.00, ],
        [ 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00,
          60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00,
          60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, 60.00, ],
        [ 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00,
          55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00,
          55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, 55.00, ],
        [ 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00,
          50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00,
          50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, 50.00, ],
        [ 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00,
          45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00,
          45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, 45.00, ],
    ], dtype=numpy.float64)
    geolat_c = geolat_c.T[:, :, numpy.newaxis, numpy.newaxis, numpy.newaxis, numpy.newaxis]

    # Make the value an approximate sphere surface area (in square degrees) of the quadrilateral
    vals  = geolon_c[:-1, :-1] * geolat_c[:-1,  1:]
    vals -= geolon_c[:-1,  1:] * geolat_c[:-1, :-1]
    vals += geolon_c[:-1,  1:] * geolat_c[ 1:,  1:]
    vals -= geolon_c[ 1:,  1:] * geolat_c[:-1,  1:]
    vals += geolon_c[ 1:,  1:] * geolat_c[ 1:, :-1]
    vals -= geolon_c[ 1:, :-1] * geolat_c[ 1:,  1:]
    vals += geolon_c[ 1:, :-1] * geolat_c[:-1, :-1]
    vals -= geolon_c[:-1, :-1] * geolat_c[ 1:, :-1]
    vals = 0.5 * numpy.fabs(vals)
    vals *= numpy.cos( 0.25 * numpy.deg2rad(geolat_c[:-1, :-1] + \
                                            geolat_c[:-1,  1:] + \
                                            geolat_c[ 1:,  1:] + \
                                            geolat_c[ 1:, :-1]) )

    # Assign the value of the rectangles between 60E and 100E with the missing value
    resbdf = numpy.array([-99999.0], dtype=numpy.float64)
    inpbdfs = numpy.array([-88888.0, -77777.0, -66666.0, -55555.0, -44444.0, -33333.0], dtype=numpy.float64)
    vals[14,:,0,0] = inpbdfs[3]

    # Make sure these calls do not generate errors
    info = ferret_init(0)
    del info
    limits = ferret_result_limits(0)
    del limits

    # Create the shapefile
    result = numpy.ones((1,1,1,1,1,1), dtype=numpy.float64)
    ferret_compute(0, result, resbdf, (shapefilename, geolon_c, geolat_c, vals, fieldname, ""), inpbdfs)

    # Read the shapefile back in and check
    sfreader = shapefile.Reader(shapefilename)
    shapes = sfreader.shapes()
    records = sfreader.records()
    explen = (vals.shape[0] - 1) * vals.shape[1]
    if len(shapes) != explen:
        raise ValueError("Expected %d shapes; found %d" % (explen, len(shapes)))
    if len(records) != explen:
        raise ValueError("Expected %d records; found %d" % (explen, len(records)))
    # Create the expected arrays of shape coordinates and values
    exppoints = []
    expvals = []
    for j in range(vals.shape[1]):
        for i in range(vals.shape[0]):
            if vals[i, j, 0, 0, 0, 0] != inpbdfs[3]:
                exppoints.append( numpy.array([ [ geolon_c[i,   j,   0, 0, 0, 0], 
                                                  geolat_c[i,   j,   0, 0, 0, 0] ],
                                                [ geolon_c[i+1, j,   0, 0, 0, 0], 
                                                  geolat_c[i+1, j,   0, 0, 0, 0] ],
                                                [ geolon_c[i+1, j+1, 0, 0, 0, 0], 
                                                  geolat_c[i+1, j+1, 0, 0, 0, 0] ],
                                                [ geolon_c[i,   j+1, 0, 0, 0, 0], 
                                                  geolat_c[i,   j+1, 0, 0, 0, 0] ],
                                                [ geolon_c[i,   j,   0, 0, 0, 0], 
                                                  geolat_c[i,   j,   0, 0, 0, 0] ] ]) )
                expvals.append(vals[i, j, 0, 0, 0, 0])
    # Verify these arrays - does not depend on the same ordering of the shapes
    for (shape, record) in zip(shapes, records):
        for k in range(len(exppoints)):
            if numpy.allclose(shape.points, exppoints[k], rtol=1.0E-4):
                break
        else:
            raise ValueError("Unexpected vertices %s" % str(shape.points))
        if not numpy.allclose(record, expvals[k], rtol=1.0E-4):
            raise ValueError("Expected value %s; found %s for shape.points %s" % \
                             (str(expvals[k]), str(record), str(shape.points)))
        junk = exppoints.pop(k)
        junk = expvals.pop(k)
    # Verify the projection file
    prjfile = file("%s.prj" % shapefilename, "r")
    datalines = prjfile.readlines()
    prjfile.close()
    if len(datalines) != 1:
        raise ValueError("Number of lines in the .prj file: expected: 1, found %d" % len(datalines))
    descript = datalines[0].strip()
    if descript != wgs84_descript:
        raise ValueError("Description in the .prj file:\n" \
                         "    expect: %s\n" \
                         "    found:  %s" % (wgs84_descript, descript))

    # Shapefile data files no longer needed
    os.remove("%s.dbf" % shapefilename)
    os.remove("%s.shp" % shapefilename)
    os.remove("%s.shx" % shapefilename)
    os.remove("%s.prj" % shapefilename)

    # Commented-out code used for further testing
    testcode = """
    sortedvals = numpy.sort(vals[ vals != inpbdfs[3] ])
    numvals = sortedvals.shape[0]
    limits = [ sortedvals[1 * numvals // 5],
               sortedvals[2 * numvals // 5],
               sortedvals[3 * numvals // 5],
               sortedvals[4 * numvals // 5], ]
    print str( [ sortedvals[0] ] + limits + [ sortedvals[-1] ] )

    partvals = vals.copy()
    partvals[ partvals >= limits[0] ] = inpbdfs[3]
    ferret_compute(0, result, resbdf, (shapefilename + "_1",
                                       geolon_c, geolat_c, partvals,
                                       fieldname, ""), inpbdfs)

    partvals = vals.copy()
    partvals[ partvals <  limits[0] ] = inpbdfs[3]
    partvals[ partvals >= limits[1] ] = inpbdfs[3]
    ferret_compute(0, result, resbdf, (shapefilename + "_2",
                                       geolon_c, geolat_c, partvals,
                                       fieldname, ""), inpbdfs)

    partvals = vals.copy()
    partvals[ partvals <  limits[1] ] = inpbdfs[3]
    partvals[ partvals >= limits[2] ] = inpbdfs[3]
    ferret_compute(0, result, resbdf, (shapefilename + "_3",
                                       geolon_c, geolat_c, partvals,
                                       fieldname, ""), inpbdfs)
    partvals = vals.copy()
    partvals[ partvals <  limits[2] ] = inpbdfs[3]
    partvals[ partvals >= limits[3] ] = inpbdfs[3]
    ferret_compute(0, result, resbdf, (shapefilename + "_4",
                                       geolon_c, geolat_c, partvals,
                                       fieldname, ""), inpbdfs)

    partvals = vals.copy()
    partvals[ partvals <  limits[3] ] = inpbdfs[3]
    ferret_compute(0, result, resbdf, (shapefilename + "_5",
                                       geolon_c, geolat_c, partvals,
                                       fieldname, ""), inpbdfs)
    """

    print("shapefile_writexyval: SUCCESS")




'''
Regridder for converting data between a curvilinear longitude,
latitude grid and a rectilinear longitude, latitude grid.
Uses the ESMP interface to ESMF to perform the regridding.

@author: Karl Smith
'''

import numpy
import ESMP


class CurvRectRegridder(object):
    '''
    Regridder for regridding data between a 2D curvilinear grid, where
    the longitude and latitude of each grid corner and/or center point
    is explicitly defined, and a 2D rectilinear grid, where the grid
    corners are all intersections of a given set of strictly increasing
    longitudes with a set of strictly increasing latitudes.

    For these grids, the center point [i, j] is taken to be the center
    point of the quadrilateral defined by connecting consecutive corner
    points in the sequence (corner_pt[i, j], corner_pt[i+1, j],
    corner_pt[i+1, j+1], corner_pt([i, j+1], corner_pt[i, j]).

    Uses the ESMP interface to ESMF to perform the regridding.  Prior
    to calling any instance methods in the CurvRectRegridder class, the
    ESMP module must be imported and ESMP.ESMP_Initialize() must have
    been called.  When a CurvRectRegridder instance is no longer needed,
    the finalize method of the instance should be called to free ESMP
    resources associated with the instance.  When ESMP is no longer
    required, the ESMP.ESMP_Finalize() method should be called to free
    all ESMP and ESMF resources.

    See the ESMPControl singleton class to simplify initializing and
    finalizing ESMP once, and only once, for a Python session.
    '''


    def __init__(self):
        '''
        Initializes to an empty regridder.  The ESMP module must be
        imported and ESMP.ESMP_Initialize() called (possibly through
        invoking ESMPControl().startCheckESMP()) prior to calling
        any methods in this instance.
        '''
        # tuples giving the shape of the grid (defined by number of cells)
        self.__curv_shape = None
        self.__rect_shape = None
        # ESMP_Grid objects describing the grids
        self.__curv_grid = None
        self.__rect_grid = None
        # ESMP_Field objects providing the source or regridded data
        self.__curv_dest_field = None
        self.__curv_src_field = None
        self.__rect_dest_field = None
        self.__rect_src_field = None
        # handles to regridding operations between the fields/grids
        self.__curv_to_rect_handles = { }
        self.__rect_to_curv_handles = { }


    def createCurvGrid(self, center_lons, center_lats, center_ignore=None,
                       corner_lons=None, corner_lats=None, corner_ignore=None):
        '''
        Create the curvilinear grid as an ESMP_Grid using the provided center
        longitudes and latitudes as the grid center points, and, if given,
        the grid corner longitudes and latitudes as the grid corner points.
        Curvilinear data is assigned to the center points.  Grid point
        coordinates are assigned as coord[i,j] = ( lons[i,j], lats[i,j] ).

        For these grids, the center point [i, j] is taken to be the center
        point of the quadrilateral defined by connecting consecutive corner
        points in the sequence (corner_pt[i, j], corner_pt[i+1, j],
        corner_pt[i+1, j+1], corner_pt([i ,j+1], corner_pt[i, j]).

        Any previous ESMP_Grid, ESMP_Field, or ESMP regridding procedures
        are destroyed.

        Arguments:
            center_lons:   2D array of longitudes, in degrees,
                           for each of the curvilinear center points
            center_lats:   2D array of latitudes, in degrees,
                           for each of the curvilinear center points
            center_ignore: 2D array of boolean-like values, indicating if
                           the corresponding grid center point should be
                           ignored in the regridding; if None, no grid
                           center points will be ignored
            corner_lons:   2D array of longitudes, in degrees,
                           for each of the curvilinear corner points
            corner_lats:   2D array of latitudes, in degrees,
                           for each of the curvilinear corner points
            corner_ignore: 2D array of boolean-like values, indicating if
                           the corresponding grid corner point should be
                           ignored in the regridding; if None, no grid
                           corner points will be ignored
        Returns:
            None
        Raises:
            ValueError: if the shape (dimensionality) of an argument is
                        invalid, or if a value in an argument is invalid
            TypeError:  if an argument is not array-like
        '''
        # Make sure center_lons is an appropriate array-like argument
        center_lons_array = numpy.array(center_lons, dtype=numpy.float64, copy=False)
        if len(center_lons_array.shape) != 2:
            raise ValueError("center_lons must be two-dimensional")

        # Make sure center_lats is an appropriate array-like argument
        center_lats_array = numpy.array(center_lats, dtype=numpy.float64, copy=False)
        if center_lats_array.shape != center_lons_array.shape:
            raise ValueError("center_lats and center_lons must have the same shape")

        if center_ignore == None:
            # Using all points, no mask created
            center_ignore_array = None
        else:
            # Make sure ignore_pts is an appropriate array-like argument
            center_ignore_array = numpy.array(center_ignore, dtype=numpy.bool, copy=False)
            if center_ignore_array.shape != center_lons_array.shape:
                raise ValueError("center_ignore and center_lons must have the same shape")
            # If not actually ignoring any points, do not create a mask
            if not center_ignore_array.any():
                center_ignore_array = None

        if (corner_lons != None) and (corner_lats != None):
            # Corner points specified

            # Make sure corner_lons is an appropriate array-like argument
            corner_lons_array = numpy.array(corner_lons, dtype=numpy.float64, copy=False)
            if corner_lons_array.shape != (center_lons_array.shape[0] + 1,
                                           center_lons_array.shape[1] + 1):
                raise ValueError("corner_lons must have one more point along " \
                                 "each dimension when compared to center_lons")

            # Make sure corner_lats is an appropriate array-like argument
            corner_lats_array = numpy.array(corner_lats, dtype=numpy.float64, copy=False)
            if corner_lats_array.shape != corner_lons_array.shape:
                raise ValueError("corner_lats and corner_lons must have the same shape")

            if corner_ignore == None:
                # Using all points, no mask created
                corner_ignore_array = None
            else:
                # Make sure ignore_pts is an appropriate array-like argument
                corner_ignore_array = numpy.array(corner_ignore, dtype=numpy.bool, copy=False)
                if corner_ignore_array.shape != corner_lons_array.shape:
                    raise ValueError("corner_ignore and corner_lons must have the same shape")
                # If not actually ignoring any points, do not create a mask
                if not corner_ignore_array.any():
                    corner_ignore_array = None

        elif corner_lons != None:
            raise ValueError("corner_lons given without corner_lats")
        elif corner_lats != None:
            raise ValueError("corner_lats given without corner_lons")
        elif corner_ignore != None:
            raise ValueError("corner_ignore given without corner_lons and corner_lats")
        else:
            # No corner points specified
            corner_lats_array = None
            corner_lons_array = None
            corner_ignore_array = None

        # If a centers mask is given and corner points are given, 
        # but a corners mask is not given, create a mask ignoring 
        # all corners around an ignored center
        if (corner_ignore_array == None) and (corner_lons_array != None) \
                and (center_ignore_array != None):
            corner_ignore_array = numpy.zeros((center_ignore_array.shape[0]+1, 
                                               center_ignore_array.shape[1]+1),
                                              dtype=numpy.bool, order='F')
            corner_ignore_array[:-1,:-1] = center_ignore_array
            corner_ignore_array[1:,:-1] = numpy.logical_or(corner_ignore_array[1:,:-1], 
                                                           center_ignore_array)
            corner_ignore_array[1:,1:] = numpy.logical_or(corner_ignore_array[1:,1:], 
                                                           center_ignore_array)
            corner_ignore_array[:-1,1:] = numpy.logical_or(corner_ignore_array[:-1,1:], 
                                                           center_ignore_array)

        # Release any regridding procedures and clear the dictionaries
        for handle in self.__rect_to_curv_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__rect_to_curv_handles.clear()
        for handle in self.__curv_to_rect_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__curv_to_rect_handles.clear()
        # Destroy any curvilinear ESMP_Fields
        if self.__curv_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_src_field)
            self.__curv_src_field = None
        if self.__curv_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_dest_field)
            self.__curv_dest_field = None
        # Destroy any previous curvilinear ESMP_Grid
        if self.__curv_grid != None:
            ESMP.ESMP_GridDestroy(self.__curv_grid);
            self.__curv_grid = None

        # Create the curvilinear longitude, latitude ESMP_Grid using
        # ESMP_GridCreateNoPeriDim for the typical case (not the
        # whole world) in Ferret.
        self.__curv_shape = center_lons_array.shape
        grid_shape = numpy.array(self.__curv_shape, dtype=numpy.int32)
        self.__curv_grid = ESMP.ESMP_GridCreateNoPeriDim(grid_shape,
                                         ESMP.ESMP_COORDSYS_SPH_DEG,
                                         ESMP.ESMP_TYPEKIND_R8)

        if corner_lons_array != None:
            # Allocate space for the grid corner coordinates
            ESMP.ESMP_GridAddCoord(self.__curv_grid, ESMP.ESMP_STAGGERLOC_CORNER)

            # Retrieve the grid corner coordinate arrays in the ESMP_Grid
            grid_lon_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 0,
                                                        ESMP.ESMP_STAGGERLOC_CORNER)
            grid_lat_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 1,
                                                        ESMP.ESMP_STAGGERLOC_CORNER)

            # Assign the longitudes and latitudes of the grid corners in the ESMP_Grid
            grid_lon_coords[:] = corner_lons_array.flatten('F')
            grid_lat_coords[:] = corner_lats_array.flatten('F')

            # Add a mask if not considering all the corner points
            if corner_ignore_array != None:
                # Allocate space for the grid corners mask
                ESMP.ESMP_GridAddItem(self.__curv_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                        ESMP.ESMP_STAGGERLOC_CORNER)
                # Retrieve the grid corners mask array in the ESMP_Grid
                ignore_mask = ESMP.ESMP_GridGetItem(self.__curv_grid,
                                                    ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CORNER)
                # Assign the mask in the ESMP_Grid; 
                # False (turns into zero) means use the point;
                # True (turns into one) means ignore the point
                ignore_mask[:] = corner_ignore_array.flatten('F')

        # Allocate space for the grid center coordinates
        ESMP.ESMP_GridAddCoord(self.__curv_grid, ESMP.ESMP_STAGGERLOC_CENTER)

        # Retrieve the grid center coordinate arrays in the ESMP_Grid
        grid_lon_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 0,
                                                    ESMP.ESMP_STAGGERLOC_CENTER)
        grid_lat_coords = ESMP.ESMP_GridGetCoordPtr(self.__curv_grid, 1,
                                                    ESMP.ESMP_STAGGERLOC_CENTER)

        # Assign the longitudes and latitudes of the grid centers in the ESMP_Grid
        grid_lon_coords[:] = center_lons_array.flatten('F')
        grid_lat_coords[:] = center_lats_array.flatten('F')

        # Add a mask if not considering all the center points
        if center_ignore_array != None:
            # Allocate space for the grid centers mask
            ESMP.ESMP_GridAddItem(self.__curv_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CENTER)
            # Retrieve the grid centers mask array in the ESMP_Grid
            ignore_mask = ESMP.ESMP_GridGetItem(self.__curv_grid,
                                                ESMP.ESMP_GRIDITEM_MASK,
                                                ESMP.ESMP_STAGGERLOC_CENTER)
            # Assign the mask in the ESMP_Grid; 
            # False (turns into zero) means use the point;
            # True (turns into one) means ignore the point
            ignore_mask[:] = center_ignore_array.flatten('F')


    def assignCurvField(self, data=None):
        '''
        Possibly creates, and possible assigns, an appropriate curvilinear
        ESMP_Field located at the center points of the grid.  An ESMP_Field
        is created if and only if it is not already present; thus allowing
        the "weights" computed from a previous regridding to be reused.

        If data is not None, assigns the data values to the curvilinear
        source ESMP_Field, creating the ESMP_field if it does not exist.

        If data is None, only creates the curvilinear destination ESMP_Field
        if it does not exist.

        Arguments:
            data: 2D array of data values to be assigned in an
                  ESMP_Field for the grid center points; if None,
                  no data in any ESMP_Field is modified.
        Returns:
            None
        Raises:
            ValueError: if data is not None, and if the shape
                        (dimensionality) of data is invalid or
                        if a value in data is not numeric
            TypeError:  if data, if not None, is not array-like
        '''
        if data == None:

            # Create the curvilinear destination ESMP_Field if it does not exist
            if self.__curv_dest_field == None:
                self.__curv_dest_field = ESMP.ESMP_FieldCreateGrid(self.__curv_grid,
                                                        "curv_dest_field",
                                                        ESMP.ESMP_TYPEKIND_R8,
                                                        ESMP.ESMP_STAGGERLOC_CENTER)

        else:

            # Make sure data is an appropriate array-like argument
            data_array = numpy.array(data, dtype=numpy.float64, copy=True)
            if data_array.shape != self.__curv_shape:
                raise ValueError("data must have the same shape " \
                                 "as the center points arrays")

            # Create the curvilinear source ESMP_Field if it does not exist
            if self.__curv_src_field == None:
                self.__curv_src_field = ESMP.ESMP_FieldCreateGrid(self.__curv_grid,
                                                       "curv_src_field",
                                                       ESMP.ESMP_TYPEKIND_R8,
                                                       ESMP.ESMP_STAGGERLOC_CENTER)

            # Retrieve the field data array in the ESMP_Field
            field_ptr = ESMP.ESMP_FieldGetPtr(self.__curv_src_field)

            # Assign the data to the field array in the ESMP_Field
            field_ptr[:] = data_array.flatten('F')


    def createRectGrid(self, center_lons, center_lats, center_ignore= None,
                       corner_lons=None, corner_lats=None, corner_ignore=None):
        '''
        Create the rectilinear grid as an ESMP_Grid using the provided center
        longitudes and latitudes as the grid center points, and, if given,
        the grid corner longitudes and latitudes as the grid corner points.
        Rectilinear data is assigned to the center points.  Grid point
        coordinates are assigned as coord[i,j] = ( lons[i], lats[j] ).

        For these grids, the center point [i,j] is taken to be the center
        point of the quadrilateral defined by connecting consecutive corner
        points in the sequence ( corner_pt[i, j], corner_pt[i+1, j],
        corner_pt[i+1, j+1], corner_pt([i, j+1], corner_pt[i, j] ).

        Any previous ESMP_Grid, ESMP_Field, or ESMP regridding procedures
        are destroyed.

        Arguments:
            center_lons:   1D array of longitudes, in degrees,
                           for each of the rectilinear grid cells
                           center planes
            center_lats:   1D array of latitudes, in degrees,
                           for each of the rectilinear grid cells
                           center planes
            center_ignore: 2D array of boolean-like values, indicating if
                           the corresponding grid center point should be
                           ignored in the regridding; if None, no grid
                           center points will be ignored
            corner_lons:   1D array of longitudes, in degrees,
                           for each of the rectilinear grid cells
                           corner planes
            corner_lats:   1D array of latitudes, in degrees,
                           for each of the rectilinear grid cells
                           corner planes
            corner_ignore: 2D array of boolean-like values, indicating if
                           the corresponding grid corner point should be
                           ignored in the regridding; if None, no grid
                           corner points will be ignored
        Returns:
            None
        Raises:
            ValueError: if the shape (dimensionality) of an argument is
                        invalid, or if a value in an argument is invalid
            TypeError:  if an argument is not array-like
        '''
        # Make sure center_lons is an appropriate array-like argument
        center_lons_array = numpy.array(center_lons, dtype=numpy.float64, copy=False)
        if len(center_lons_array.shape) != 1:
            raise ValueError("center_lons must be one-dimensional")

        # Make sure center_lats is an appropriate array-like argument
        center_lats_array = numpy.array(center_lats, dtype=numpy.float64, copy=True)
        if len(center_lats_array.shape) != 1:
            raise ValueError("center_lats must be one-dimensional")

        if center_ignore == None:
            # Using all center points, no mask created
            center_ignore_array = None
        else:
            # Make sure center_ignore is an appropriate array-like argument
            center_ignore_array = numpy.array(center_ignore, dtype=numpy.bool, copy=False)
            if center_ignore_array.shape != (center_lons_array.shape[0],
                                             center_lats_array.shape[0]):
                raise ValueError("center_ignore must have the shape " \
                                 "( len(center_lons), len(center_lats) )")
            # If not actually ignoring any center points, do not create a mask
            if not center_ignore_array.any():
                center_ignore_array = None

        if (corner_lons != None) and (corner_lats != None):
            # Corner points specified

            # Make sure corner_lons is an appropriate array-like argument
            corner_lons_array = numpy.array(corner_lons, dtype=numpy.float64, copy=False)
            if corner_lons_array.shape != (center_lons_array.shape[0] + 1, ):
                raise ValueError("corner_lons must have shape ( len(center_lons) + 1, )")

            # Make sure corner_lats is an appropriate array-like argument
            corner_lats_array = numpy.array(corner_lats, dtype=numpy.float64, copy=True)
            if corner_lats_array.shape != (center_lats_array.shape[0] + 1, ):
                raise ValueError("corner_lats must have shape ( len(center_lats) + 1, )")

            if corner_ignore == None:
                # Using all corner points, no mask created
                corner_ignore_array = None
            else:
                # Make sure corner_ignore is an appropriate array-like argument
                corner_ignore_array = numpy.array(corner_ignore, dtype=numpy.bool, copy=False)
                if corner_ignore_array.shape != (corner_lons_array.shape[0],
                                                 corner_lats_array.shape[0]):
                    raise ValueError("corner_ignore must have the shape " \
                                     "( len(corner_lons), len(corner_lats) )")
                # If not actually ignoring any points, do not create a mask
                if not corner_ignore_array.any():
                    corner_ignore_array = None

        elif corner_lons != None:
            raise ValueError("corner_lons given without corner_lats")
        elif corner_lats != None:
            raise ValueError("corner_lats given without corner_lons")
        elif corner_ignore != None:
            raise ValueError("corner_ignore given without corner_lons and corner_lats")
        else:
            # No corner points specified
            corner_lats_array = None
            corner_lons_array = None
            corner_ignore_array = None

        # If a centers mask is given and corner points are given, 
        # but a corners mask is not given, create a mask ignoring 
        # all corners around an ignored center
        if (corner_ignore_array == None) and (corner_lons_array != None) \
                and (center_ignore_array != None):
            corner_ignore_array = numpy.zeros((center_ignore_array.shape[0]+1, 
                                               center_ignore_array.shape[1]+1),
                                              dtype=numpy.bool, order='F')
            corner_ignore_array[:-1,:-1] = center_ignore_array
            corner_ignore_array[1:,:-1] = numpy.logical_or(corner_ignore_array[1:,:-1], 
                                                           center_ignore_array)
            corner_ignore_array[1:,1:] = numpy.logical_or(corner_ignore_array[1:,1:], 
                                                           center_ignore_array)
            corner_ignore_array[:-1,1:] = numpy.logical_or(corner_ignore_array[:-1,1:], 
                                                           center_ignore_array)

        # Release any regridding procedures and clear the dictionaries
        for handle in self.__rect_to_curv_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__rect_to_curv_handles.clear()
        for handle in self.__curv_to_rect_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__curv_to_rect_handles.clear()
        # Destroy any rectilinear ESMP_Fields
        if self.__rect_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_src_field)
            self.__rect_src_field = None
        if self.__rect_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_dest_field)
            self.__rect_dest_field = None
        # Destroy any previous rectilinear ESMP_Grid
        if self.__rect_grid != None:
            ESMP.ESMP_GridDestroy(self.__rect_grid);
            self.__rect_grid = None

        # Create the rectilinear longitude, latitude grid using
        # ESMP_GridCreateNoPeriDim for the typical case (not the
        # whole world) in Ferret.
        self.__rect_shape = (center_lons_array.shape[0],
                             center_lats_array.shape[0])
        grid_shape = numpy.array(self.__rect_shape, dtype=numpy.int32)
        self.__rect_grid = ESMP.ESMP_GridCreateNoPeriDim(grid_shape,
                                         ESMP.ESMP_COORDSYS_SPH_DEG,
                                         ESMP.ESMP_TYPEKIND_R8)

        if corner_lons_array != None:
            # Allocate space for the grid corner coordinates
            ESMP.ESMP_GridAddCoord(self.__rect_grid, ESMP.ESMP_STAGGERLOC_CORNER)

            # Retrieve the grid corner coordinate arrays in the ESMP_Grid
            grid_lon_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 0,
                                                        ESMP.ESMP_STAGGERLOC_CORNER)
            grid_lat_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 1,
                                                        ESMP.ESMP_STAGGERLOC_CORNER)

            # Assign the longitudes and latitudes of the grid corners in the ESMP_Grid
            # numpy.tile([a,b,c], 2) = [a,b,c,a,b,c]
            grid_lon_coords[:] = numpy.tile(corner_lons_array, corner_lats_array.shape[0])
            # numpy.repeat([a,b,c], 2) = [a,a,b,b,c,c]
            grid_lat_coords[:] = numpy.repeat(corner_lats_array, corner_lons_array.shape[0])

            # Add a mask if not considering all the corner points
            if corner_ignore_array != None:
                # Allocate space for the grid corners mask
                ESMP.ESMP_GridAddItem(self.__rect_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                        ESMP.ESMP_STAGGERLOC_CORNER)
                # Retrieve the grid corners mask array in the ESMP_Grid
                ignore_mask = ESMP.ESMP_GridGetItem(self.__rect_grid,
                                                    ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CORNER)
                # Assign the mask in the ESMP_Grid; 
                # False (turns into zero) means use the point;
                # True (turns into one) means ignore the point;
                # flatten in column-major (F) order to match lon-lat assignment order
                ignore_mask[:] = corner_ignore_array.flatten('F')

        # Allocate space for the grid center coordinates
        ESMP.ESMP_GridAddCoord(self.__rect_grid, ESMP.ESMP_STAGGERLOC_CENTER)

        # Retrieve the grid corner coordinate arrays in the ESMP_Grid
        grid_lon_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 0,
                                                    ESMP.ESMP_STAGGERLOC_CENTER)
        grid_lat_coords = ESMP.ESMP_GridGetCoordPtr(self.__rect_grid, 1,
                                                    ESMP.ESMP_STAGGERLOC_CENTER)

        # Assign the longitudes and latitudes of the grid centers in the ESMP_Grid
        # numpy.tile([a,b,c], 2) = [a,b,c,a,b,c]
        grid_lon_coords[:] = numpy.tile(center_lons_array, center_lats_array.shape[0])
        # numpy.repeat([a,b,c], 2) = [a,a,b,b,c,c]
        grid_lat_coords[:] = numpy.repeat(center_lats_array, center_lons_array.shape[0])

        # Add a mask if not considering all the center points
        if center_ignore_array != None:
            # Allocate space for the grid centers mask
            ESMP.ESMP_GridAddItem(self.__rect_grid, ESMP.ESMP_GRIDITEM_MASK,
                                                    ESMP.ESMP_STAGGERLOC_CENTER)
            # Retrieve the grid centers mask array in the ESMP_Grid
            ignore_mask = ESMP.ESMP_GridGetItem(self.__rect_grid,
                                                ESMP.ESMP_GRIDITEM_MASK,
                                                ESMP.ESMP_STAGGERLOC_CENTER)
            # Assign the mask in the ESMP_Grid; 
            # False (turns into zero) means use the point;
            # True (turns into one) means ignore the point;
            # flatten in column-major (F) order to match lon-lat assignment order
            ignore_mask[:] = center_ignore_array.flatten('F')


    def assignRectField(self, data=None):
        '''
        Possibly creates, and possible assigns, an appropriate rectilinear
        ESMP_Field located at the center points of the grid.  An ESMP_Field
        is created if and only if it is not already present; thus allowing
        the "weights" computed from a previous regridding to be reused.

        If data is not None, assigns the data values to the rectilinear
        source ESMP_Field, creating the ESMP_field if it does not exist.

        If data is None, only creates the rectilinear destination ESMP_Field
        if it does not exist.

        Arguments:
            data: 2D array of data values to be assigned in an
                  ESMP_Field for the grid center points; if None,
                  no data in any ESMP_Field is modified.
        Returns:
            None
        Raises:
            ValueError: if data is not None, and if the shape
                        (dimensionality) of data is invalid or
                        if a value in data is not numeric
            TypeError:  if data, if not None, is not array-like
        '''
        if data == None:

            # Create the rectilinear destination ESMP_Field if it does not exist
            if self.__rect_dest_field == None:
                self.__rect_dest_field = ESMP.ESMP_FieldCreateGrid(self.__rect_grid,
                                                        "rect_dest_field",
                                                        ESMP.ESMP_TYPEKIND_R8,
                                                        ESMP.ESMP_STAGGERLOC_CENTER)

        else:

            # Make sure data is an appropriate array-like argument
            data_array = numpy.array(data, dtype=numpy.float64, copy=True)
            if len(data_array.shape) != 2:
                raise ValueError("data must be two-dimensional")
            if data_array.shape != self.__rect_shape:
                raise ValueError("data must have the same shape " \
                                 "as the center points arrays")

            # Create the rectilinear source ESMP_Field if it does not exist
            if self.__rect_src_field == None:
                self.__rect_src_field = ESMP.ESMP_FieldCreateGrid(self.__rect_grid,
                                                       "rect_src_field",
                                                       ESMP.ESMP_TYPEKIND_R8,
                                                       ESMP.ESMP_STAGGERLOC_CENTER)

            # Retrieve the field data array in the ESMP_Field
            field_ptr = ESMP.ESMP_FieldGetPtr(self.__rect_src_field)

            # Assign the data to the field array in the ESMP_Field
            field_ptr[:] = data_array.flatten('F')


    def regridCurvToRect(self, undef_val,
                         method=ESMP.ESMP_REGRIDMETHOD_BILINEAR):
        '''
        Regrids from the curvilinear source ESMP_Field to the rectilinear
        destination ESMP_Field using the given regridding method.  Reuses
        the appropriate regridding procedure if one already exists;
        otherwise a new regridding procedure is created and stored.

        Prior to calling this method, the curvilinear source ESMP_Field
        must be created by calling createCurvGrid, then assignCurvField
        with valid data.  The rectilinear destination ESMP_Field must
        also have been created by calling createRectGrid, and then
        assignRectField with no data argument (or None for data).

        Arguments:
            undef_val: numpy array containing one numeric value to
                       be used as the undefined data value in the
                       returned array
            method:    one of the ESMP regridding method identifiers,
                       such as:
                           ESMP.ESMP_REGRIDMETHOD_BILINEAR
                           ESMP.ESMP_REGRIDMETHOD_CONSERVE
                           ESMP.ESMP_REGRIDMETHOD_PATCH
                       Conservative regridding requires that both
                       corner and center point coordinates are
                       defined in the grids.
        Returns:
            data: a 2D numpy array of data values located at the rectilinear
                  grid centers representing the regridded curvilinear
                  ESMP_Field data.  The undefined data value will be
                  assigned to unassigned data points.
        Raises:
            ValueError: if either the curvilinear source ESMP_Field or
                        the rectilinear destination ESMP_Field does not
                        exist.
        '''
        # Check that the source and destination fields exist
        if self.__curv_src_field == None:
            raise ValueError("Curvilinear source ESMP_Field does not exist")
        if self.__rect_dest_field == None:
            raise ValueError("Rectilinear destination ESMP_Field does not exist")
        # Check if a regrid procedure handle already exists for this method
        handle = self.__curv_to_rect_handles.get(method, None)
        # If no handle found, create one
        if handle == None:
            # Assign the value in the masks marking points to be ignored
            ignore_mask_value = numpy.array([1], dtype=numpy.int32)
            # Generate the procedure handle
            handle = ESMP.ESMP_FieldRegridStore(self.__curv_src_field,
                                                self.__rect_dest_field,
                                                ignore_mask_value, ignore_mask_value,
                                                method, ESMP.ESMP_UNMAPPEDACTION_IGNORE)
            # Save the handle for this method for future regrids
            self.__curv_to_rect_handles[method] = handle
        # Initialize the destination field values with the undefined data value
        field_ptr = ESMP.ESMP_FieldGetPtr(self.__rect_dest_field)
        field_ptr[:] = undef_val
        # Perform the regridding, zeroing out only the 
        # destination fields values that will be assigned
        ESMP.ESMP_FieldRegrid(self.__curv_src_field, self.__rect_dest_field,
                              handle, ESMP.ESMP_REGION_SELECT)
        # Make a copy of the destination field values to return, reshaped to 2D
        result = numpy.array(field_ptr, dtype=numpy.float64, copy=True)
        result = result.reshape(self.__rect_shape, order='F')

        return result


    def regridRectToCurv(self, undef_val,
                         method=ESMP.ESMP_REGRIDMETHOD_BILINEAR):
        '''
        Regrids from the rectilinear source ESMP_Field to the curvilinear
        destination ESMP_Field using the given regridding method.  Reuses
        the appropriate regridding procedure if one already exists;
        otherwise a new regridding procedure is created and stored.

        Prior to calling this method, the rectilinear source ESMP_Field
        must be created by calling createRectGrid, then assignRectField
        with valid data.  The curvilinear destination ESMP_Field must
        also have been created by calling createCurvGrid, and then
        assignCurvField with no data argument (or None for data).

        Arguments:
            undef_val: numpy array containing one numeric value to
                       be used as the undefined data value in the
                       returned array
            method:    one of the ESMP regridding method identifiers,
                       such as:
                           ESMP.ESMP_REGRIDMETHOD_BILINEAR
                           ESMP.ESMP_REGRIDMETHOD_CONSERVE
                           ESMP.ESMP_REGRIDMETHOD_PATCH
                       Conservative regridding requires that both
                       corner and center point coordinates are
                       defined in the grids.
        Returns:
            data: a 2D array of data values located at the curvilinear
                  grid centers representing the regridded rectilinear
                  ESMP_Field data.  The undefined data value will be
                  assigned to unassigned data points.
        Raises:
            ValueError: if either the rectilinear source ESMP_Field or
                        the curvilinear destination ESMP_Field does not
                        exist.
        '''
        # Check that the source and destination fields exist
        if self.__rect_src_field == None:
            raise ValueError("Rectilinear source ESMP_Field does not exist")
        if self.__curv_dest_field == None:
            raise ValueError("Curvilinear destination ESMP_Field does not exist")
        # Check if a regrid procedure handle already exists for this method
        handle = self.__rect_to_curv_handles.get(method, None)
        # If no handle found, create one
        if handle == None:
            # Assign the value in the masks marking points to be ignored
            ignore_mask_value = numpy.array([1], dtype=numpy.int32)
            # Generate the procedure handle
            handle = ESMP.ESMP_FieldRegridStore(self.__rect_src_field,
                                                self.__curv_dest_field,
                                                ignore_mask_value, ignore_mask_value,
                                                method, ESMP.ESMP_UNMAPPEDACTION_IGNORE)
            # Save the handle for this method for future regrids
            self.__rect_to_curv_handles[method] = handle
        # Initialize the destination field values with the undefined data value
        field_ptr = ESMP.ESMP_FieldGetPtr(self.__curv_dest_field)
        field_ptr[:] = undef_val
        # Perform the regridding, zeroing out only the 
        # destination fields values that will be assigned
        ESMP.ESMP_FieldRegrid(self.__rect_src_field, self.__curv_dest_field,
                              handle, ESMP.ESMP_REGION_SELECT)
        # Make a copy of the destination field values to return, reshaped to 2D
        result = numpy.array(field_ptr, dtype=numpy.float64, copy=True)
        result = result.reshape(self.__curv_shape, order='F')

        return result


    def finalize(self):
        '''
        Destroys any ESMP_Grid, ESMP_Field, and ESMP regridding
        procedures present in this instance.  If ESMP is no longer
        needed, ESMP.ESMP_Finalize() (possibly by invoking
        ESMPControl().stopESMP()) should be called to free any
        ESMP and ESMF resources.

        Arguments:
            None
        Returns:
            None
        '''
        # Release any regridding procedures and clear the dictionaries
        for handle in self.__rect_to_curv_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__rect_to_curv_handles.clear()
        for handle in self.__curv_to_rect_handles.values():
            ESMP.ESMP_FieldRegridRelease(handle)
        self.__curv_to_rect_handles.clear()
        # Destroy any ESMP_Fields
        if self.__rect_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_src_field)
            self.__rect_src_field = None
        if self.__rect_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__rect_dest_field)
            self.__rect_dest_field = None
        if self.__curv_src_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_src_field)
            self.__curv_src_field = None
        if self.__curv_dest_field != None:
            ESMP.ESMP_FieldDestroy(self.__curv_dest_field)
            self.__curv_dest_field = None
        # Destroy any ESMP_Grids
        if self.__rect_grid != None:
            ESMP.ESMP_GridDestroy(self.__rect_grid);
            self.__rect_grid = None
        if self.__curv_grid != None:
            ESMP.ESMP_GridDestroy(self.__curv_grid);
            self.__curv_grid = None
        # clear the shape attributes just to be complete
        self.__rect_shape = None
        self.__curv_shape = None





''' Represents a forecast-model-run collection in Ferret '''

import pyferret
from pyferret.ferfmrcvar import _fmrc_var_qualifier

class FerFMRCDSet(pyferret.FerAggDSet):
    '''
    A forecast-model-run collection dataset in Ferret.  Variables in this dataset
    (FerFMRCVar objects) have restrictions on the time and forecast axes, and have
    special transformations to better view the quality of the forecast represented 
    by the variable.
    '''

    def __init__(self, name, dsets, title='', warn=True, hide=False):
        '''
        Creates a forecast-model-run collection dataset in Ferret.  Variables in 
        the given datasets to be aggregated must have a time axes (forecasted time)
        whose values are offset but otherwise match where they overlap (a subset
        of thimes that advances along a single time axis that is the union of all 
        dataset time axes).  The datasets can be given in any order; they will be 
        arranged to have monotonically increasing time subsets.  The aggregated
        variables (FerFMRCVar objects) have special transformations to better view 
        the quality of the forecast represented by this variable.
        '''
        super(FerFMRCDSet, self).__init__(name=name, dsets=dsets, along='F', 
                                          title=title, warn=warn, hide=hide)
        # The datasets have been aggregated along the F axis and all the known 
        # variables added to this dataset's _fervars dictionary.  Turn each of 
        # these FerVar objects into a FerFMRCVar object.
        for uppername in self._fervars.keys():
            if uppername == 'TF_TIMES':
                continue
            self._fervars[uppername] = pyferret.FerFMRCVar(self._fervars[uppername], 
                                                           _fmrc_var_qualifier)

    def __repr__(self):
        '''
        Representation to of this FerFMRCDSet.
        Includes the variable names as variables can be added after creation.
        '''
        infostr = "FerFMRCDSet(name='%s', dsets=%s, hide=%s) with variables %s" % \
                  (self._dsetname, str(self._compdsetnames), 
                   str(self._comphidden), str(self.fernames(sort=True)))
        return infostr

    # Note: Just use the __eq__ and __ne__ methods inherited from FerAggDSet.
    #       No need to make these objects appear to be different from a FerAggDSet.




'''
Represents an aggregation of data sets
'''

from __future__ import print_function

import sys
import os
import tempfile
import pyferret
from pyferret.ferdset import _anonymous_dataset_qualifier

class FerAggDSet(pyferret.FerDSet):
    '''
    An aggregation of data sets and the variables they have in common
    '''

    def __init__(self, name, dsets, along='T', title='', warn=True, hide=False):
        '''
        Aggregates the given list of datasets along the given axis using the 
        Ferret "DEFINE DATA /AGGREGATE" command.  Creates a FerVar for each data 
        variable in common among these datasets, and assigns it as an attribute 
        of this class instance using the variable name.  
            name (string): Ferret name for this aggregated dataset
            dsets (sequence of strings and/or FerDSets): datasets to aggregate.
                A string will be interpreted as a filename for creating a FerDSet.
            along ('T', 'E', 'F'): axis along which to aggregate the datasets
            title (string): title for the dataset for plots and listing;
                if not given, the Ferret name for the dataset will be used
            warn (bool): issue warning messages about variables not in common among
                all member datasets (either not present or not using the same grid)
            hide (bool): hide the member datasets in standard Ferret listings 
                such as with pyferret.showdata()
        '''
        # Create an empty dataset with the given Ferret name
        super(FerAggDSet, self).__init__('', qual=_anonymous_dataset_qualifier)
        if not isinstance(name, str):
            raise ValueError('Ferret name for the aggregate dataset must be astring')
        aggname = name.strip()
        if not aggname:
            raise ValueError('Ferret name for the aggregate dataset is blank')
        self._filename = aggname
        self._dsetname = aggname
        # Need to keep the given order of component datasets
        self._compdsetnames = [ ]
        # But still use a dictionary with uppercase names for keys
        self._compdsets = { }

        if along not in ('T', 'E', 'F'):
            raise ValueError("along must be one of 'T', 'E', or 'F'")
        self._along = along
        self._comphidden = bool(hide)

        # Create a Ferret string variable containing all the dataset names to be aggregated
        if not ( isinstance(dsets, tuple) or isinstance(dsets, list) ):
            raise ValueError('dsets must be a tuple or list of strings and/or FerDSets')
        filesfile = tempfile.NamedTemporaryFile(mode='w', delete=False, 
                                  prefix=aggname + '_', suffix='_agg.txt')
        filesfilename = filesfile.name
        deletefilesfile = True
        try:
            for myitem in dsets:
                if isinstance(myitem, str):
                    mydset = pyferret.FerDSet(myitem)
                elif isinstance(myitem, pyferret.FerDSet):
                    mydset = myitem
                else:
                    raise ValueError('dsets must be a tuple or list of strings and/or FerDSets')
                if mydset._dsetname.upper() in self._compdsets:
                    raise ValueError('duplicate dataset name ' + mydset._dsetname)
                print(mydset._dsetname, file=filesfile)
                self._compdsetnames.append(mydset._dsetname)
                self._compdsets[mydset._dsetname.upper()] = mydset
            deletefilesfile = False
        finally:
            filesfile.close()
            if deletefilesfile:
                os.unlink(filesfilename)
        filesvarname = aggname + "_datafile_names"
        cmdstr = 'LET ' + filesvarname + ' = SPAWN("cat \'' + filesfilename + '\'")'
        (errval, errmsg) = pyferret.run(cmdstr)
        if errval != pyferret.FERR_OK:
            os.unlink(filesfilename)
            raise ValueError(errmsg)
        # filesfile not read (SPAWN command executed) until filesvarname is needed

        # Create the DEFINE DATA /AGGREGATE Ferret command, creating
        # and saving component FerDSets as needed
        cmdstr = 'DEFINE DATA/AGGREGATE/' + self._along
        if title:
            cmdstr += '/TITLE="' + str(title) + '"'
        if not warn:
            cmdstr += '/QUIET'
        if self._comphidden:
            cmdstr += '/HIDE'
        cmdstr += ' ' + aggname + ' = ' + filesvarname
        (errval, errmsg) = pyferret.run(cmdstr)
        # filesfile now read so can delete it
        os.unlink(filesfilename)
        if errval != pyferret.FERR_OK:
            raise ValueError(errmsg)

        # create a FerVar for each variable in this dataset
        namesdict = pyferret.getstrdata('..varnames')
        for varname in namesdict['data'].flatten():
            if sys.version_info[0] > 2:
                # For Python3.x, namesdict['data'] is a NumPy array of bytes; convert to unicode
                varname = str(varname, 'UTF-8')
            # create a FerVar representing this existing Ferret aggregated file variable
            filevar = pyferret.FerVar()
            filevar._markasknownvar(varname, self._dsetname, True)
            # assign this FerVar - uppercase the variable name keys to make case-insensitive
            self._fervars[varname.upper()] = filevar
            # keep a original-case version of the name
            self._fervarnames.add(varname)


    def __repr__(self):
        '''
        Representation to of this FerAggDSet.
        Includes the variable names as variables can be added after creation.
        '''
        infostr = "FerAggDSet(name='%s', dsets=%s, along='%s', hide=%s) with variables %s" % \
                  (self._dsetname, str(self._compdsetnames), self._along, 
                   str(self._comphidden), str(self.fernames(sort=True)))
        return infostr


    def __eq__(self, other):
        '''
        Two FerAggDSets are equal if their Ferret names, lists of aggregated 
        dataset names, aggregation axis, component dataset hidden status, and 
        dictionary of FerVar are variables all equal.  All string values are 
        compared case-insensitive.
        '''
        if not isinstance(other, pyferret.FerDSet):
            return NotImplemented
        if not isinstance(other, FerAggDSet):
            return False
        if not super(FerAggDSet, self).__eq__(other):
            return False
        if self._along != other._along:
            return False
        if self._comphidden != other._comphidden:
            return False
        if len(self._compdsetnames) != len(other._compdsetnames):
            return False
        for k in range(len(self._compdsetnames)):
            if self._compdsetnames[k].upper() != other._compdsetnames[k].upper():
                return False
        return True


    def __ne__(self, other):
        '''
        Two FerDSets are not equal if their Ferret names, lists of aggregated
        dataset names, or dictionary of FerVar variables are not equal.
        All string values are compared case-insensitive.
        '''
        if not isinstance(other, pyferret.FerDSet):
            return NotImplemented
        return not self.__eq__(other)


    def getdsetnames(self):
        '''
        Returns a copy of the list of component dataset names (original-case)
        in the order of their aggregation.
        '''
        return list(self._compdsetnames)


    def getdsets(self):
        '''
        Returns a list of component FerDSet datasets 
        in the order of their aggregation.
        '''
        return [ self._compdsets[name.upper()] for name in self._compdsetnames ]


    def close(self):
        '''
        Removes (cancels) all the variables in Ferret associated with this dataset,
        then closes (cancels) this dataset in Ferret.  If the aggregated dataset was 
        created with hide=True, this will close (cancel) all the component datasets 
        as well; otherwise the component datasets and dataset names will remain here
        and in Ferret.  Raises a ValueError if there is a problem.
        '''
        # if the dataset is already closed, ignore this command
        if not self._dsetname:
            return
        # run the Ferret CANCEL commands in FerDSet.close
        super(FerAggDSet, self).close()
        # if component datasets were hidden, the above close also closed all the
        # component datasets, so also clear the component dataset information
        if self._comphidden:
            self._compdsets.clear()
            self._compdsetnames = [ ]




"""
Returns the array of probability distribution function values
for the Normal probability distribution
using the given arrays for the abscissa or template
values and each of the parameters values.
"""

from __future__ import print_function

import numpy
import pyferret
import pyferret.stats
import scipy.stats

DISTRIB_NAME = "Normal"
FUNC_NAME = "pdf"


def ferret_init(id):
    """
    Initialization for the stats_norm_pdf Ferret PyEF
    """
    return pyferret.stats.getinitdict(DISTRIB_NAME, FUNC_NAME)


def ferret_compute(id, result, resbdf, inputs, inpbdfs):
    """
    Result array assignment for the stats_norm_pdf Ferret PyEF
    """
    pyferret.stats.assignresultsarray(DISTRIB_NAME, FUNC_NAME,
                                      result, resbdf, inputs, inpbdfs)


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # make sure ferret_init does not have problems
    info = ferret_init(0)

    # Normal distribution in the YZ plane, MUs on the X axis, SIGMAs on the T
    xdim = 6
    ydim = 25
    zdim = 15
    tdim = 4
    yzvals = numpy.linspace(0.0, 100.0, ydim * zdim)
    mus = numpy.linspace(20.0, 45.0, xdim)
    sigmas = numpy.linspace(8.0, 14.0, tdim)
    pdfs = numpy.empty((xdim, ydim, zdim, tdim, 1, 1), dtype=numpy.float64, order='F')
    for i in range(xdim):
        for q in range(tdim):
            distf = scipy.stats.norm(mus[i], sigmas[q])
            values = distf.pdf(yzvals)
            pdfs[i, :, :, q, 0, 0] = values.reshape((ydim, zdim), order='F')
    # configure arrays for ferret_compute
    yzvals = numpy.array(yzvals, dtype=numpy.float64).reshape((1, ydim, zdim, 1, 1, 1), order='F')
    mus = numpy.array(mus, dtype=numpy.float64).reshape((xdim, 1, 1, 1, 1, 1), order='F')
    sigmas = numpy.array(sigmas, dtype=numpy.float64).reshape((1, 1, 1 , tdim, 1, 1), order='F')
    inpbdfs = numpy.array([-9999.0, -8888.0, -7777.0], dtype=numpy.float64)
    resbdf = numpy.array([-6666.0], dtype=numpy.float64)
    # Throw in some undefined values
    index = 0
    for k in range(zdim):
        for j in range(ydim):
            if (index % 13) == 3:
                abscissa[0, j, k, 0, 0, 0] = inpbdfs[0]
                pdfs[:, j, k, :, 0, 0] = resbdf[0]
    mus[4, 0, 0, 0, 0, 0] = inpbdfs[1]
    pdfs[4, :, :, :, 0, 0] = resbdf[0]
    sigmas[0, 0, 0, 1, 0, 0] = inpbdfs[2]
    pdfs[:, :, :, 1, 0, 0] = resbdf[0]
    # Get the result from ferret_compute and compare
    result = -5555.0 * numpy.ones((xdim, ydim, zdim, tdim, 1, 1), dtype=numpy.float64, order='F')
    ferret_compute(0, result, resbdf, (yzvals, mus, sigmas), inpbdfs)
    if not numpy.allclose(result, pdfs):
        print("Expect (flattened) =\n%s" % str(pdfs.reshape(-1)))
        print("Result (flattened) =\n%s" % str(result.reshape(-1)))
        raise ValueError("Unexpected result")

    # All successful
    print("Success")




# To be run in python after importing and starting pyferret
# such as from running "pyferret -python"

from __future__ import print_function

import numpy

import sys ; sys.ps1 = '' ; sys.ps2 = ''
print()

print(">>> emptyregion = pyferret.FerRegion()")
emptyregion = pyferret.FerRegion()
print(">>> print repr(emptyregion)")
print(repr(emptyregion))
print(">>> emptyregion._ferretqualifierstr()")
emptyregion._ferretqualifierstr()

print(">>> coordregion = pyferret.FerRegion(X=slice(-70,20),Y=slice('20N','60N'),Z=-100,T='15-DEC-2015',E=0,F='1')")
coordregion = pyferret.FerRegion(X=slice(-70,20),Y=slice('20N','60N'),Z=-100,T='15-DEC-2015',E=0,F='1')
print(">>> print repr(coordregion)")
print(repr(coordregion))
print(">>> coordregion._ferretqualifierstr()")
coordregion._ferretqualifierstr()

print(">>> indexregion = pyferret.FerRegion(I=slice(0,20),J=slice('1','60'),K=10,L='5',M=0,N='1')")
indexregion = pyferret.FerRegion(I=slice(0,20),J=slice('1','60'),K=10,L='5',M=0,N='1')
print(">>> print repr(indexregion)")
print(repr(indexregion))
print(">>> indexregion._ferretqualifierstr()")
indexregion._ferretqualifierstr()

print(">>> try:")
print("...     badregion = pyferret.FerRegion(X='70W:20E',I=slice(0,20))")
print("...     print 'Error not caught'")
print("... except ValueError as ex:")
print("...     print 'Error caught: %s' % str(ex)")
try:
    badregion = pyferret.FerRegion(X='70W:20E',I=slice(0,20))
    print('Error not caught')
except ValueError as ex:
    print('Error caught: %s' % str(ex))

